@@ -33,69 +33,6 @@
 
 #include "pocl_types.h"
 
-#if (__clang_major__ == 3)
-# if (__clang_minor__ == 7)
-# undef LLVM_3_7
-# define LLVM_3_7
-#elif (__clang_minor__ == 8)
-# undef LLVM_3_8
-# define LLVM_3_8
-#elif (__clang_minor__ == 9)
-# undef LLVM_3_9
-# define LLVM_3_9
-#endif
-
-#elif (__clang_major__ == 4)
-
-# undef LLVM_4_0
-# define LLVM_4_0
-
-#elif (__clang_major__ == 5)
-
-# undef LLVM_5_0
-# define LLVM_5_0
-
-#elif (__clang_major__ == 6)
-
-# undef LLVM_6_0
-# define LLVM_6_0
-
-#else
-
-#error Unsupported Clang/LLVM version.
-
-#endif
-
-#ifndef LLVM_6_0
-#define LLVM_OLDER_THAN_6_0 1
-
-#ifndef LLVM_5_0
-#define LLVM_OLDER_THAN_5_0 1
-
-#ifndef LLVM_4_0
-#define LLVM_OLDER_THAN_4_0 1
-
-#ifndef LLVM_3_9
-#define LLVM_OLDER_THAN_3_9 1
-
-#ifndef LLVM_3_8
-#define LLVM_OLDER_THAN_3_8 1
-
-#ifndef LLVM_3_7
-#define LLVM_OLDER_THAN_3_7 1
-
-#ifndef LLVM_3_6
-#define LLVM_OLDER_THAN_3_6 1
-
-#endif
-#endif
-#endif
-#endif
-#endif
-#endif
-#endif
-
-
 #include "_kernel_constants.h"
 
 /* Function/type attributes supported by Clang/SPIR */
@@ -114,21 +51,22 @@
 #else
 #  define _CL_OVERLOADABLE
 #endif
-#if __has_attribute(__const__)
-#  define _CL_READNONE __attribute__((__const__))
-#else
-#  define _CL_READNONE
-#endif
 #if __has_attribute(__pure__)
 #  define _CL_READONLY __attribute__((__pure__))
 #else
 #  define _CL_READONLY
 #endif
-#if __has_attribute(__unavailable__)
-#  define _CL_UNAVAILABLE __attribute__((__unavailable__))
+#if __has_attribute(__const__)
+#  define _CL_READNONE __attribute__((__const__))
 #else
-#  define _CL_UNAVAILABLE
+#  define _CL_READNONE
 #endif
+#if __has_attribute(convergent)
+#  define _CL_CONVERGENT __attribute__((convergent))
+#else
+#  define _CL_CONVERGENT
+#endif
+
 
 typedef char char2  __attribute__((__ext_vector_type__(2)));
 typedef char char3  __attribute__((__ext_vector_type__(3)));
@@ -217,7 +155,7 @@ typedef ulong ulong16 __attribute__((__ext_vector_type__(16)));
    the default builtins we use C functions which require
    the typedefs to the actual underlying types.
 */
-#if defined(__CBUILD__) && defined(CLANG_OLDER_THAN_3_9)
+#if defined(__CBUILD__) && defined(LLVM_OLDER_THAN_3_9)
 typedef int sampler_t;
 
 /* Since some built-ins have different return types
@@ -257,15 +195,39 @@ typedef struct _pocl_image1d_array_t { dev_image_t base; }* image1d_array_t;
 #elif defined(__TCE__)
 
 #define POCL_ADDRESS_SPACE_PRIVATE 0
-#define POCL_ADDRESS_SPACE_GLOBAL 3
-#define POCL_ADDRESS_SPACE_LOCAL 4
-#define POCL_ADDRESS_SPACE_CONSTANT 5
+#define POCL_ADDRESS_SPACE_GLOBAL 1
+#define POCL_ADDRESS_SPACE_LOCAL 3
+#define POCL_ADDRESS_SPACE_CONSTANT 2
 #define POCL_ADDRESS_SPACE_GENERIC 6
 
 #endif
 
 typedef uint cl_mem_fence_flags;
 
-#include "_enable_all_exts.h"
+/* Integer Constants */
+
+#if defined(__CBUILD__)
+
+#define CHAR_BIT  8
+#define CHAR_MAX  SCHAR_MAX
+#define CHAR_MIN  SCHAR_MIN
+#define INT_MAX   2147483647
+#define INT_MIN   (-2147483647 - 1)
+#ifdef cl_khr_int64
+#define LONG_MAX  0x7fffffffffffffffL
+#define LONG_MIN  (-0x7fffffffffffffffL - 1)
+#endif
+#define SCHAR_MAX 127
+#define SCHAR_MIN (-127 - 1)
+#define SHRT_MAX  32767
+#define SHRT_MIN  (-32767 - 1)
+#define UCHAR_MAX 255
+#define USHRT_MAX 65535
+#define UINT_MAX  0xffffffff
+#ifdef cl_khr_int64
+#define ULONG_MAX 0xffffffffffffffffUL
+#endif
+
+#endif /* __CBUILD__ */
 
 #endif
