@@ -82,8 +82,14 @@ poclu_create_any_context();
  * All input parameters must be allocated by caller!
  * Returns CL_SUCCESS on success, or a descriptive OpenCL error code upon failure.
  */
-POCLU_API cl_int POCLU_CALL
-poclu_get_any_device( cl_context *context, cl_device_id *device, cl_command_queue *queue);
+POCLU_API cl_int POCLU_CALL poclu_get_any_device2 (cl_context *context,
+                                                   cl_device_id *device,
+                                                   cl_command_queue *queue,
+                                                   cl_platform_id *platform);
+
+POCLU_API cl_int POCLU_CALL poclu_get_any_device (cl_context *context,
+                                                  cl_device_id *device,
+                                                  cl_command_queue *queue);
 
 /**
  * cl_half related helpers.
@@ -105,18 +111,21 @@ poclu_read_binfile(char *filename, size_t *len);
 POCLU_API int POCLU_CALL
 poclu_write_file(char* filemane, char* content, size_t size);
 
-
+int poclu_load_program (cl_context context, cl_device_id device,
+                        const char *basename, int spir, int spirv, cl_program *p);
 /* In case cl_err != CL_SUCCESS, prints out the error + function : line to stderr,
  * and returns 1, otherwise returns 0
  */
 POCLU_API int POCLU_CALL
 check_cl_error(cl_int cl_err, int line, const char* func_name);
 
-#define _POCLU_CHECK_CL_ERROR_INNER(cond, func, line)                   \
-do {                                                                    \
-   if(check_cl_error(cond, line, func))                                 \
-     exit(EXIT_FAILURE);                                                \
-} while (0)
+#define _POCLU_CHECK_CL_ERROR_INNER(cond, func, line)                         \
+  do                                                                          \
+    {                                                                         \
+      if (check_cl_error (cond, line, func))                                  \
+        return (EXIT_FAILURE);                                                \
+    }                                                                         \
+  while (0)
 
 #define CHECK_CL_ERROR(cond) _POCLU_CHECK_CL_ERROR_INNER(cond, __PRETTY_FUNCTION__, __LINE__)
 
@@ -132,6 +141,10 @@ do {                                                                    \
   }                                                                     \
 } while (0)
 
+#define CHECK_CL_ERROR2(err)                                                  \
+  if (check_cl_error (err, __LINE__, __PRETTY_FUNCTION__))                    \
+  goto ERROR
+
 #ifdef __cplusplus
 }
 #endif
