@@ -25,6 +25,7 @@
 #include "pocl_cl.h"
 #include "devices.h"
 #include "common.h"
+#include "pocl_util.h"
 
 CL_API_ENTRY cl_mem CL_API_CALL
 POname(clCreateBuffer)(cl_context   context,
@@ -115,7 +116,6 @@ POname(clCreateBuffer)(cl_context   context,
   mem->type = CL_MEM_OBJECT_BUFFER;
   mem->flags = flags;
   mem->is_image = CL_FALSE;
-  mem->latest_event = NULL;
   mem->owning_device = NULL;
   mem->is_pipe = 0;
   mem->pipe_packet_size = 0;
@@ -127,11 +127,7 @@ POname(clCreateBuffer)(cl_context   context,
   mem->device_ptrs =
     (pocl_mem_identifier*) calloc(pocl_num_devices,
                                   sizeof(pocl_mem_identifier));
-  if (mem->device_ptrs == NULL)
-    {
-      errcode = CL_OUT_OF_HOST_MEMORY;
-      goto ERROR;
-    }
+  POCL_GOTO_ERROR_COND((mem->device_ptrs == NULL), CL_OUT_OF_HOST_MEMORY);
 
   /* init dev pointer structure to ease inter device pointer sharing
      in ops->alloc_mem_obj */
@@ -180,7 +176,7 @@ POname(clCreateBuffer)(cl_context   context,
   if ((flags & CL_MEM_ALLOC_HOST_PTR) && (mem->mem_host_ptr == NULL))
     {
       assert(mem->shared_mem_allocation_owner == NULL);
-      mem->mem_host_ptr = pocl_memalign_alloc (MAX_EXTENDED_ALIGNMENT, size);
+      mem->mem_host_ptr = pocl_aligned_malloc (MAX_EXTENDED_ALIGNMENT, size);
       if (mem->mem_host_ptr == NULL)
         {
           errcode = CL_OUT_OF_HOST_MEMORY;
@@ -191,8 +187,8 @@ POname(clCreateBuffer)(cl_context   context,
   POCL_RETAIN_OBJECT(context);
 
   POCL_MSG_PRINT_MEMORY (
-      "Created Buffer %p, HOST_PTR: %p, DEVICE_PTR[0]: %p \n", mem,
-      mem->mem_host_ptr, mem->device_ptrs[0].mem_ptr);
+      "Created Buffer %p, HOST_PTR: %p, DEVICE_PTR[0]: %p SIZE %zu \n", mem,
+      mem->mem_host_ptr, mem->device_ptrs[0].mem_ptr, size);
 
   if (errcode_ret != NULL)
     *errcode_ret = CL_SUCCESS;
