@@ -25,6 +25,7 @@
 #include "pocl_cl.h"
 #include "pocl_util.h"
 #include "pocl_mem_management.h"
+#include "pocl_shared.h"
 #include <stdlib.h>
 #include <string.h>
 
@@ -115,6 +116,9 @@ int context_set_properties(cl_context                    context,
     }
 }
 
+unsigned cl_context_count = 0;
+pocl_lock_t pocl_context_handling_lock = POCL_LOCK_INITIALIZER;
+
 CL_API_ENTRY cl_context CL_API_CALL
 POname(clCreateContext)(const cl_context_properties * properties,
                 cl_uint                       num_devices,
@@ -125,16 +129,31 @@ POname(clCreateContext)(const cl_context_properties * properties,
 {
   unsigned i;
   cl_device_id device_ptr;
-  int errcode = 0;
+  cl_int errcode = 0;
   cl_context context = NULL;
 
+  POCL_LOCK (pocl_context_handling_lock);
+
   POCL_GOTO_ERROR_COND((devices == NULL || num_devices == 0), CL_INVALID_VALUE);
 
   POCL_GOTO_ERROR_COND((pfn_notify == NULL && user_data != NULL), CL_INVALID_VALUE);
 
-  
+  int offline_compile = pocl_get_bool_option("POCL_OFFLINE_COMPILE", 0);
+
   lt_dlinit();
-  pocl_init_devices();
+  errcode = pocl_init_devices();
+  /* clCreateContext cannot return CL_DEVICE_NOT_FOUND, which is what
+   * pocl_init_devices() returns if no devices could be probed. Hence,
+   * remap this error to CL_INVALID_DEVICE. Note that this particular
+   * situation should never arise, since an application should issue
+   * clGetDeviceIDs before clCreateContext, and we would have returned
+   * CL_DEVICE_NOT_FOUND already from clGetDeviceIDs. Still, no reason
+   * not to handle it.
+   */
+  if (errcode == CL_DEVICE_NOT_FOUND)
+    errcode = CL_INVALID_DEVICE;
+  POCL_GOTO_ERROR_ON ((errcode != CL_SUCCESS), errcode,
+                      "Could not initialize devices\n");
 
   context = (cl_context) malloc(sizeof(struct _cl_context));
   if (context == NULL)
@@ -170,9 +189,14 @@ POname(clCreateContext)(const cl_context_properties * properties,
           goto ERROR_CLEAN_CONTEXT_AND_DEVICES;
         }
       
-      if (device_ptr->available != CL_TRUE)
+      if (!offline_compile && (device_ptr->available != CL_TRUE))
         {
-          POCL_MSG_WARN("Dropping unavailable device: %s\n", device_ptr->long_name);
+          POCL_MSG_WARN("Offline compilation disabled - dropping unavailable device: %s\n", device_ptr->long_name);
+          context->devices[i] = context->devices[--context->num_devices];
+        }
+      else if((device_ptr->compiler_available != CL_TRUE) && (device_ptr->available != CL_TRUE))
+        {
+          POCL_MSG_WARN("Device unavailable and device compiler unavailable - dropping device: %s\n", device_ptr->long_name);
           context->devices[i] = context->devices[--context->num_devices];
         }
       else
@@ -193,6 +217,10 @@ POname(clCreateContext)(const cl_context_properties * properties,
   if (errcode_ret)
     *errcode_ret = CL_SUCCESS;
   context->valid = 1;
+
+  cl_context_count += 1;
+  POCL_UNLOCK (pocl_context_handling_lock);
+
   return context;
   
 ERROR_CLEAN_CONTEXT_AND_DEVICES:
@@ -205,6 +233,8 @@ ERROR:
     {
       *errcode_ret = errcode;
     }
+
+  POCL_UNLOCK (pocl_context_handling_lock);
   return NULL;
 }
 POsym(clCreateContext)
