@@ -1,8 +1,9 @@
 /* OpenCL runtime library: clEnqueueCopyBuffer()
 
-   Copyright (c) 2011-2012 Universidad Rey Juan Carlos and
+   Copyright (c) 2011-2015 Universidad Rey Juan Carlos and
                            Pekka Jääskeläinen / Tampere Univ. of Tech.
-   
+                           Ville Korhonen / Tampere Univ. of Tech.
+
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
@@ -42,6 +43,7 @@ CL_API_SUFFIX__VERSION_1_0
   unsigned i;
   _cl_command_node *cmd = NULL;
   int errcode;
+  cl_mem buffers[2] = {src_buffer, dst_buffer};
 
   POCL_RETURN_ERROR_COND((command_queue == NULL), CL_INVALID_COMMAND_QUEUE);
 
@@ -60,11 +62,10 @@ CL_API_SUFFIX__VERSION_1_0
 
   POCL_RETURN_ERROR_COND((size == 0), CL_INVALID_VALUE);
 
-  POCL_RETURN_ERROR_COND((event_wait_list == NULL && num_events_in_wait_list > 0),
-    CL_INVALID_EVENT_WAIT_LIST);
-
-  POCL_RETURN_ERROR_COND((event_wait_list != NULL && num_events_in_wait_list == 0),
-    CL_INVALID_EVENT_WAIT_LIST);
+  errcode = pocl_check_event_wait_list (command_queue, num_events_in_wait_list,
+                                        event_wait_list);
+  if (errcode != CL_SUCCESS)
+    return errcode;
 
   if (pocl_buffers_boundcheck(src_buffer, dst_buffer, src_offset,
         dst_offset, size) != CL_SUCCESS) return CL_INVALID_VALUE;
@@ -76,28 +77,22 @@ CL_API_SUFFIX__VERSION_1_0
 
   errcode = pocl_create_command (&cmd, command_queue, CL_COMMAND_COPY_BUFFER, 
                                  event, num_events_in_wait_list, 
-                                 event_wait_list);
+                                 event_wait_list, 2, buffers);
   if (errcode != CL_SUCCESS)
     return errcode;
 
-  cmd->command.copy.src_buffer = src_buffer;
-  cmd->command.copy.dst_buffer = dst_buffer;
-  
-  POname(clRetainMemObject)(src_buffer);
-  POname(clRetainMemObject)(dst_buffer);
-
-  cmd->command.copy.data = device->data;
-  /* TODO: call device->buf_offset() or similar as device_ptrs might not be
-     actual buffer pointers but pointers to a book keeping structure. */
-  cmd->command.copy.src_ptr = 
-    (char*)src_buffer->device_ptrs[device->dev_id].mem_ptr;
+  cmd->command.copy.src_mem_id = &src_buffer->device_ptrs[device->dev_id];
   cmd->command.copy.src_offset = src_offset;
 
-  cmd->command.copy.dst_ptr = 
-    (char*)dst_buffer->device_ptrs[device->dev_id].mem_ptr;
+  cmd->command.copy.dst_mem_id = &dst_buffer->device_ptrs[device->dev_id];
   cmd->command.copy.dst_offset = dst_offset;
-  cmd->command.copy.cb = size;
+  cmd->command.copy.size = size;
 
+  POname(clRetainMemObject)(src_buffer);
+  src_buffer->owning_device = command_queue->device;
+  POname(clRetainMemObject)(dst_buffer);
+  dst_buffer->owning_device = command_queue->device;
+  
   pocl_command_enqueue(command_queue, cmd);
 
   return CL_SUCCESS;
