@@ -86,9 +86,8 @@ CL_API_SUFFIX__VERSION_1_2
   if (errcode != CL_SUCCESS)
     return errcode;
 
-  cmd->command.memfill.buffer = buffer;
-  cmd->command.memfill.ptr =
-      buffer->device_ptrs[command_queue->device->dev_id].mem_ptr;
+  cmd->command.memfill.dst_mem_id
+      = &buffer->device_ptrs[command_queue->device->dev_id];
   cmd->command.memfill.size = size;
   cmd->command.memfill.offset = offset;
   void *p = pocl_aligned_malloc(pattern_size, pattern_size);
@@ -98,7 +97,7 @@ CL_API_SUFFIX__VERSION_1_2
 
   POname(clRetainMemObject) (buffer);
   buffer->owning_device = command_queue->device;
-  
+
   pocl_command_enqueue(command_queue, cmd);
 
   return CL_SUCCESS;
