@@ -46,6 +46,7 @@ POname(clEnqueueReadBufferRect)(cl_command_queue command_queue,
   cl_device_id device;
   unsigned i;
   _cl_command_node *cmd;
+  int errcode;
 
   POCL_RETURN_ERROR_COND((command_queue == NULL), CL_INVALID_COMMAND_QUEUE);
 
@@ -61,11 +62,10 @@ POname(clEnqueueReadBufferRect)(cl_command_queue command_queue,
   POCL_RETURN_ERROR_ON((command_queue->context != buffer->context),
     CL_INVALID_CONTEXT, "buffer and command_queue are not from the same context\n");
 
-  POCL_RETURN_ERROR_COND((event_wait_list == NULL && num_events_in_wait_list > 0),
-    CL_INVALID_EVENT_WAIT_LIST);
-
-  POCL_RETURN_ERROR_COND((event_wait_list != NULL && num_events_in_wait_list == 0),
-    CL_INVALID_EVENT_WAIT_LIST);
+  errcode = pocl_check_event_wait_list (command_queue, num_events_in_wait_list,
+                                        event_wait_list);
+  if (errcode != CL_SUCCESS)
+    return errcode;
 
   POCL_RETURN_ERROR_COND((ptr == NULL), CL_INVALID_VALUE);
   POCL_RETURN_ERROR_COND((buffer_origin == NULL), CL_INVALID_VALUE);
@@ -92,25 +92,31 @@ POname(clEnqueueReadBufferRect)(cl_command_queue command_queue,
                       (unsigned long)buffer_row_pitch, (unsigned long)buffer_slice_pitch, 
                       (unsigned long)host_row_pitch, (unsigned long)host_slice_pitch);
   
-  POname(clRetainMemObject) (buffer);
-  
   pocl_create_command (&cmd, command_queue, CL_COMMAND_READ_BUFFER_RECT,
                        event, num_events_in_wait_list, event_wait_list, 1, 
                        &buffer);
 
-  cmd->command.read_image.device_ptr = 
-    buffer->device_ptrs[device->dev_id].mem_ptr;
-  cmd->command.read_image.host_ptr = ptr;
-  memcpy (&cmd->command.read_image.origin, buffer_origin, sizeof (size_t) * 3);
-  memcpy (&cmd->command.read_image.h_origin, host_origin, sizeof (size_t) * 3);
-  memcpy (&cmd->command.read_image.region, region, sizeof (size_t) * 3);
-  cmd->command.read_image.h_rowpitch = host_row_pitch;
-  cmd->command.read_image.h_slicepitch = host_slice_pitch;
-  cmd->command.read_image.b_rowpitch = buffer_row_pitch;
-  cmd->command.read_image.b_slicepitch = buffer_slice_pitch;
-  cmd->command.read_image.buffer = buffer;
+  cmd->command.read_rect.src_mem_id = &buffer->device_ptrs[device->dev_id];
+  cmd->command.read_rect.dst_host_ptr = ptr;
+
+  cmd->command.read_rect.host_origin[0] = host_origin[0];
+  cmd->command.read_rect.host_origin[1] = host_origin[1];
+  cmd->command.read_rect.host_origin[2] = host_origin[2];
+  cmd->command.read_rect.buffer_origin[0] = buffer_origin[0];
+  cmd->command.read_rect.buffer_origin[1] = buffer_origin[1];
+  cmd->command.read_rect.buffer_origin[2] = buffer_origin[2];
+  cmd->command.read_rect.region[0] = region[0];
+  cmd->command.read_rect.region[1] = region[1];
+  cmd->command.read_rect.region[2] = region[2];
+
+  cmd->command.read_rect.host_row_pitch = host_row_pitch;
+  cmd->command.read_rect.host_slice_pitch = host_slice_pitch;
+  cmd->command.read_rect.buffer_row_pitch = buffer_row_pitch;
+  cmd->command.read_rect.buffer_slice_pitch = buffer_slice_pitch;
+
+  POname (clRetainMemObject) (buffer);
+  buffer->owning_device = device;
 
-  buffer->owning_device = command_queue->device;
   pocl_command_enqueue (command_queue, cmd);
 
   if (blocking_read)
