@@ -33,12 +33,34 @@ POname(clReleaseEvent)(cl_event event) CL_API_SUFFIX__VERSION_1_0
   POCL_RETURN_ERROR_COND((event->context == NULL), CL_INVALID_EVENT);
 
   POCL_RELEASE_OBJECT (event, new_refcount);
-
+  
   if (new_refcount == 0)
     {
+      event_callback_item *cb_ptr = NULL;
+      event_callback_item *next = NULL;
+      for (cb_ptr = event->callback_list; cb_ptr; cb_ptr = next)
+        {
+          next = cb_ptr->next;
+          POCL_MEM_FREE (cb_ptr);
+        }
+
+      if (event->command_type == CL_COMMAND_USER)
+        {
+          pocl_user_event_data *p = event->data;
+          pthread_cond_destroy (&p->wakeup_cond);
+          POCL_MEM_FREE (p);
+        }
+
+      POCL_MSG_PRINT_REFCOUNTS ("Free event %d\n", event->id);
+      if (event->command_type != CL_COMMAND_USER &&
+          event->queue->device->ops->free_event_data)
+        event->queue->device->ops->free_event_data(event);
+
       POname(clReleaseContext) (event->context);
       if (event->queue)
         POname(clReleaseCommandQueue) (event->queue);
+
+      POCL_DESTROY_OBJECT (event);
       pocl_mem_manager_free_event (event);
     }
 
