@@ -23,9 +23,9 @@
    THE SOFTWARE.
 */
 
-/* for posix_memalign */
+/* for posix_memalign and strdup */
 #define _DEFAULT_SOURCE
-#define _POSIX_C_SOURCE 200112L
+#define _POSIX_C_SOURCE 200809L
 
 #include "common.h"
 #include "pocl_shared.h"
@@ -61,6 +61,9 @@
 
 #include "_kernel_constants.h"
 
+#if defined(__x86_64__) || defined(__i386__)
+#define CPU_IS_X86 1
+#endif
 
 #define WORKGROUP_STRING_LENGTH 1024
 
@@ -79,6 +82,7 @@ char*
 llvm_codegen (const char* tmpdir, cl_kernel kernel, cl_device_id device,
               size_t local_x, size_t local_y, size_t local_z)
 {
+  POCL_MEASURE_START (llvm_codegen);
   int error = 0;
   void *llvm_module = NULL;
 
@@ -180,24 +184,24 @@ llvm_codegen (const char* tmpdir, cl_kernel kernel, cl_device_id device,
   /* temporary filename for kernel.so */
   if (pocl_cache_tempname (tmp_module, ".so", NULL))
     {
-      POCL_MSG_PRINT_LLVM ("Creating temporary kernel.so file "
+      POCL_MSG_PRINT_LLVM ("Creating temporary kernel.so file"
                            " for kernel %s FAILED\n",
                            kernel->name);
       goto FINISH;
     }
   else
-    POCL_MSG_PRINT_LLVM ("Temporary kernel.so file "
+    POCL_MSG_PRINT_LLVM ("Temporary kernel.so file"
                          " for kernel %s : %s\n",
                          kernel->name, tmp_module);
 
   POCL_MSG_PRINT_INFO ("Linking final module\n");
   char *const args1[]
 #ifndef POCL_ANDROID
-/* on non-X86, use Clang because we need it for compiler-rt builtins library*/
-#if defined(__x86_64__) || defined(__i386__)
-      = { LINK_COMMAND,
-#else
+/* use Clang as linker, when compiler-rt is available and needed */
+#if defined(ENABLE_POCL_FLOAT_CONVERSION) || defined(CLANG_HAS_RTLIB)
       = { CLANG,
+#else
+      = { LINK_COMMAND,
 #endif
           "-o",
           tmp_module,
@@ -247,6 +251,7 @@ llvm_codegen (const char* tmpdir, cl_kernel kernel, cl_device_id device,
 FINISH:
   pocl_destroy_llvm_module (llvm_module);
   POCL_MEM_FREE (objfile);
+  POCL_MEASURE_FINISH (llvm_codegen);
 
   if (error)
     return NULL;
@@ -265,20 +270,20 @@ fill_dev_image_t (dev_image_t* di, struct pocl_argument* parg,
                   cl_device_id device)
 {
   cl_mem mem = *(cl_mem *)parg->value;
-  di->width = mem->image_width;
-  di->height = mem->image_height;
-  di->depth = mem->image_depth;
-  di->row_pitch = mem->image_row_pitch;
-  di->slice_pitch = mem->image_slice_pitch;
-  di->order = mem->image_channel_order;
-  di->image_array_size = mem->image_array_size;
-  di->data_type = mem->image_channel_data_type;
+  di->_width = mem->image_width;
+  di->_height = mem->image_height;
+  di->_depth = mem->image_depth;
+  di->_row_pitch = mem->image_row_pitch;
+  di->_slice_pitch = mem->image_slice_pitch;
+  di->_order = mem->image_channel_order;
+  di->_image_array_size = mem->image_array_size;
+  di->_data_type = mem->image_channel_data_type;
   pocl_get_image_information (mem->image_channel_order,
-                              mem->image_channel_data_type, &(di->num_channels),
-                              &(di->elem_size));
+                              mem->image_channel_data_type,
+                              &(di->_num_channels), &(di->_elem_size));
 
-  HANDLE_IMAGE1D_BUFFER (mem);
-  di->data = (mem->device_ptrs[device->dev_id].mem_ptr);
+  IMAGE1D_TO_BUFFER (mem);
+  di->_data = (mem->device_ptrs[device->dev_id].mem_ptr);
 }
 
 void
@@ -295,8 +300,11 @@ pocl_copy_mem_object (cl_device_id dest_dev, cl_mem dest,
     {
       src_dev->ops->copy 
         (dest_dev->data, 
-         source->device_ptrs[src_dev->dev_id].mem_ptr, source_offset,
-         dest->device_ptrs[dest_dev->dev_id].mem_ptr, dest_offset, 
+         &dest->device_ptrs[dest_dev->dev_id],
+         dest,
+         &source->device_ptrs[src_dev->dev_id],
+         source,
+         dest_offset, source_offset,
          cb);
     }
   else
@@ -315,11 +323,13 @@ pocl_copy_mem_object (cl_device_id dest_dev, cl_mem dest,
       
       src_dev->ops->read 
         (src_dev->data, tmp, 
-         source->device_ptrs[src_dev->dev_id].mem_ptr, source_offset, 
-         cb);
+          &source->device_ptrs[src_dev->dev_id],
+          source,
+          source_offset, cb);
       dest_dev->ops->write 
         (dest_dev->data, tmp, 
-         dest->device_ptrs[dest_dev->dev_id].mem_ptr, dest_offset,
+         &dest->device_ptrs[dest_dev->dev_id],
+          dest, dest_offset,
          cb);
       free (tofree);
     }
@@ -358,13 +368,6 @@ pocl_ndrange_node_cleanup(_cl_command_node *node)
   POname(clReleaseKernel)(node->command.run.kernel);
 }
 
-void
-pocl_native_kernel_cleanup(_cl_command_node *node)
-{
-  free (node->command.native.mem_list);
-  free (node->command.native.args);
-}
-
 void
 pocl_mem_objs_cleanup (cl_event event)
 {
@@ -372,19 +375,14 @@ pocl_mem_objs_cleanup (cl_event event)
   for (i = 0; i < event->num_buffers; ++i)
     {
       assert(event->mem_objs[i] != NULL);
-      POCL_LOCK_OBJ (event->mem_objs[i]);
-      if (event->mem_objs[i]->latest_event == event)
-        event->mem_objs[i]->latest_event = NULL;
-      POCL_UNLOCK_OBJ (event->mem_objs[i]);
       POname(clReleaseMemObject) (event->mem_objs[i]);
     }
   free (event->mem_objs);
   event->mem_objs = NULL;
 }
 
-static const size_t zero_origin[] = { 0, 0, 0 };
 /**
- * executes given command.
+ * executes given command. Call with node->event UNLOCKED.
  */
 void
 pocl_exec_command (_cl_command_node * volatile node)
@@ -392,312 +390,367 @@ pocl_exec_command (_cl_command_node * volatile node)
   unsigned i;
   /* because of POCL_UPDATE_EVENT_ */
   cl_event event = node->event;
+  cl_device_id dev = node->device;
+  _cl_command_t *cmd = &node->command;
   switch (node->type)
     {
     case CL_COMMAND_READ_BUFFER:
       POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->read
-        (node->device->data, 
-         node->command.read.host_ptr, 
-         node->command.read.device_ptr,
-         node->command.read.offset,
-         node->command.read.cb);
+      assert (dev->ops->read);
+      dev->ops->read
+        (dev->data,
+         cmd->read.dst_host_ptr,
+         cmd->read.src_mem_id,
+         event->mem_objs[0],
+         cmd->read.offset,
+         cmd->read.size);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Read Buffer           ");
       break;
+
     case CL_COMMAND_WRITE_BUFFER:
       POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->write
-        (node->device->data, 
-         node->command.write.host_ptr, 
-         node->command.write.device_ptr,
-         node->command.write.offset, 
-         node->command.write.cb);
+      assert (dev->ops->write);
+      dev->ops->write
+        (dev->data,
+         cmd->write.src_host_ptr,
+         cmd->write.dst_mem_id,
+         event->mem_objs[0],
+         cmd->write.offset,
+         cmd->write.size);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Write Buffer          ");
       break;
+
     case CL_COMMAND_COPY_BUFFER:
       POCL_UPDATE_EVENT_RUNNING(event);
-      pocl_copy_mem_object (node->command.copy.dst_dev, 
-                            node->command.copy.dst_buffer,
-                            node->command.copy.dst_offset,
-                            node->command.copy.src_dev,
-                            node->command.copy.src_buffer,
-                            node->command.copy.src_offset, 
-                            node->command.copy.cb);
+      assert (dev->ops->copy);
+      dev->ops->copy
+        (dev->data,
+         cmd->copy.dst_mem_id,
+         event->mem_objs[1],
+         cmd->copy.src_mem_id,
+         event->mem_objs[0],
+         cmd->copy.dst_offset,
+         cmd->copy.src_offset,
+         cmd->copy.size);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Copy Buffer           ");
       break;
-    case CL_COMMAND_MIGRATE_MEM_OBJECTS:
+
+    case CL_COMMAND_FILL_BUFFER:
       POCL_UPDATE_EVENT_RUNNING(event);
+      assert (dev->ops->memfill);
+      dev->ops->memfill
+        (dev->data,
+         cmd->memfill.dst_mem_id,
+         event->mem_objs[0],
+         cmd->memfill.size,
+         cmd->memfill.offset,
+         cmd->memfill.pattern,
+         cmd->memfill.pattern_size);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Fill Buffer           ");
+      pocl_aligned_free (cmd->memfill.pattern);
+      break;
+
+    case CL_COMMAND_READ_BUFFER_RECT:
+      POCL_UPDATE_EVENT_RUNNING(event);
+      assert (dev->ops->read_rect);
+      dev->ops->read_rect
+        (dev->data,
+         cmd->read_rect.dst_host_ptr,
+         cmd->read_rect.src_mem_id,
+         event->mem_objs[0],
+         cmd->read_rect.buffer_origin,
+         cmd->read_rect.host_origin,
+         cmd->read_rect.region,
+         cmd->read_rect.buffer_row_pitch,
+         cmd->read_rect.buffer_slice_pitch,
+         cmd->read_rect.host_row_pitch,
+         cmd->read_rect.host_slice_pitch);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Read Buffer Rect      ");
+      break;
+
+    case CL_COMMAND_COPY_BUFFER_RECT:
+      POCL_UPDATE_EVENT_RUNNING(event);
+      assert (dev->ops->copy_rect);
+      dev->ops->copy_rect
+        (dev->data,
+         cmd->copy_rect.dst_mem_id,
+         event->mem_objs[1],
+         cmd->copy_rect.src_mem_id,
+         event->mem_objs[0],
+         cmd->copy_rect.dst_origin,
+         cmd->copy_rect.src_origin,
+         cmd->copy_rect.region,
+         cmd->copy_rect.dst_row_pitch,
+         cmd->copy_rect.dst_slice_pitch,
+         cmd->copy_rect.src_row_pitch,
+         cmd->copy_rect.src_slice_pitch);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Copy Buffer Rect      ");
+      break;
+
+    case CL_COMMAND_WRITE_BUFFER_RECT:
+      POCL_UPDATE_EVENT_RUNNING(event);
+      assert (dev->ops->write_rect);
+      dev->ops->write_rect
+        (dev->data,
+         cmd->write_rect.src_host_ptr,
+         cmd->write_rect.dst_mem_id,
+         event->mem_objs[0],
+         cmd->write_rect.buffer_origin,
+         cmd->write_rect.host_origin,
+         cmd->write_rect.region,
+         cmd->write_rect.buffer_row_pitch,
+         cmd->write_rect.buffer_slice_pitch,
+         cmd->write_rect.host_row_pitch,
+         cmd->write_rect.host_slice_pitch);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Write Buffer Rect     ");
+      break;
+
+    case CL_COMMAND_MIGRATE_MEM_OBJECTS:
+      POCL_UPDATE_EVENT_RUNNING (event);
       pocl_migrate_mem_objects (node);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Migrate Buffer        ");
       break;
-    case CL_COMMAND_MAP_IMAGE:
-    case CL_COMMAND_MAP_BUFFER: 
-      POCL_UPDATE_EVENT_RUNNING(event);
-      POCL_LOCK_OBJ (node->command.map.buffer);
-      if (node->device->ops->map_mem != NULL)
-        node->device->ops->map_mem (node->device->data,
-                                    (node->command.map.buffer)
-                                        ->device_ptrs[node->device->dev_id]
-                                        .mem_ptr,
-                                    (node->command.map.mapping)->offset,
-                                    (node->command.map.mapping)->size,
-                                    (node->command.map.mapping)->host_ptr);
-      (node->command.map.buffer)->map_count++;
-      POCL_UNLOCK_OBJ (node->command.map.buffer);
-      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Map Image/Buffer      ");
+
+    case CL_COMMAND_MAP_BUFFER:
+      POCL_UPDATE_EVENT_RUNNING (event);
+      assert (dev->ops->map_mem);
+      POCL_LOCK_OBJ (event->mem_objs[0]);
+        dev->ops->map_mem (dev->data,
+                           cmd->map.mem_id,
+                           event->mem_objs[0],
+                           cmd->map.mapping);
+      (event->mem_objs[0])->map_count++;
+      POCL_UNLOCK_OBJ (event->mem_objs[0]);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Map Buffer            ");
       break;
-    case CL_COMMAND_WRITE_IMAGE:
-      POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->write_rect (
-         node->device->data,
-         node->command.write_image.host_ptr,
-         node->command.write_image.device_ptr,
-         node->command.write_image.origin,
-         zero_origin,
-         node->command.write_image.region,
-         node->command.write_image.b_rowpitch,
-         node->command.write_image.b_slicepitch,
-         node->command.write_image.h_rowpitch,
-         node->command.write_image.h_slicepitch);
-      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Write Image           ");
-      break;
-    case CL_COMMAND_WRITE_BUFFER_RECT:
-      POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->write_rect
-        (node->device->data,
-         node->command.write_image.host_ptr,
-         node->command.write_image.device_ptr,
-         node->command.write_image.origin,
-         node->command.write_image.h_origin,
-         node->command.write_image.region,
-         node->command.write_image.b_rowpitch,
-         node->command.write_image.b_slicepitch,
-         node->command.write_image.h_rowpitch,
-         node->command.write_image.h_slicepitch);
-      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Write Image           ");
+
+    case CL_COMMAND_COPY_IMAGE_TO_BUFFER:
+      POCL_UPDATE_EVENT_RUNNING (event);
+      assert (dev->ops->read_image_rect);
+      dev->ops->read_image_rect (
+          dev->data,
+          event->mem_objs[0],
+          cmd->read_image.src_mem_id,
+          NULL,
+          cmd->read_image.dst_mem_id,
+          cmd->read_image.origin,
+          cmd->read_image.region,
+          cmd->read_image.dst_row_pitch,
+          cmd->read_image.dst_slice_pitch,
+          cmd->read_image.dst_offset);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "CopyImageToBuffer       ");
       break;
+
     case CL_COMMAND_READ_IMAGE:
       POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->read_rect (
-         node->device->data,
-         node->command.read_image.host_ptr,
-         node->command.read_image.device_ptr,
-         node->command.read_image.origin,
-         zero_origin,
-         node->command.read_image.region,
-         node->command.read_image.b_rowpitch,
-         node->command.read_image.b_slicepitch,
-         node->command.read_image.h_rowpitch,
-         node->command.read_image.h_slicepitch);
+      assert (dev->ops->read_image_rect);
+      dev->ops->read_image_rect (
+          dev->data,
+          event->mem_objs[0],
+          cmd->read_image.src_mem_id,
+          cmd->read_image.dst_host_ptr,
+          NULL,
+          cmd->read_image.origin,
+          cmd->read_image.region,
+          cmd->read_image.dst_row_pitch,
+          cmd->read_image.dst_slice_pitch,
+          cmd->read_image.dst_offset);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Read Image            ");
       break;
-    case CL_COMMAND_READ_BUFFER_RECT:
-      POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->read_rect
-        (node->device->data, node->command.read_image.host_ptr,
-         node->command.read_image.device_ptr,
-         node->command.read_image.origin,
-         node->command.read_image.h_origin,
-         node->command.read_image.region,
-         node->command.read_image.b_rowpitch,
-         node->command.read_image.b_slicepitch,
-         node->command.read_image.h_rowpitch,
-         node->command.read_image.h_slicepitch);
-      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Read Buffer Rect      ");
-      break;
-    case CL_COMMAND_COPY_BUFFER_RECT:
+
     case CL_COMMAND_COPY_BUFFER_TO_IMAGE:
-    case CL_COMMAND_COPY_IMAGE_TO_BUFFER:
-    case CL_COMMAND_COPY_IMAGE:
       POCL_UPDATE_EVENT_RUNNING(event);
-      cl_device_id src_dev = node->command.copy_image.src_device;
-      cl_mem src_buf = node->command.copy_image.src_buffer;
-      cl_device_id dst_dev = node->command.copy_image.dst_device;
-      cl_mem dst_buf = node->command.copy_image.dst_buffer;
-
-      /* if source and destination are in the same global mem  */
-      if (src_dev->global_mem_id == dst_dev->global_mem_id)
-        {
-          node->device->ops->copy_rect
-            (node->device->data,
-             src_buf->device_ptrs[src_dev->dev_id].mem_ptr,
-             dst_buf->device_ptrs[dst_dev->dev_id].mem_ptr,
-             node->command.copy_image.src_origin,
-             node->command.copy_image.dst_origin,
-             node->command.copy_image.region,
-             node->command.copy_image.src_rowpitch,
-             node->command.copy_image.src_slicepitch,
-             node->command.copy_image.dst_rowpitch,
-             node->command.copy_image.dst_slicepitch);
-        }
-      /* if source and destination are in different global mem
-         data needs to be read to the host memory and then written to
-         destination device */
-      else
-        {
-          size_t size = node->command.copy_image.region[0]
-            * node->command.copy_image.region[1]
-            * node->command.copy_image.region[2];
-          void *tmp = malloc (size);
-
-          /* origin and slice pitch for tmp buffer */
-          const size_t null_origin[3] = {0, 0, 0};
-          size_t tmp_rowpitch =
-            node->command.copy_image.region[0];
-          size_t tmp_slicepitch =
-            node->command.copy_image.region[0]
-            * node->command.copy_image.region[1];
-
-          src_dev->ops->read_rect
-            (src_dev->data,
-             tmp,
-             src_buf->device_ptrs[src_dev->dev_id].mem_ptr,
-             node->command.copy_image.src_origin,
-             null_origin,
-             node->command.copy_image.region,
-             node->command.copy_image.src_rowpitch,
-             node->command.copy_image.src_slicepitch,
-             tmp_rowpitch,
-             tmp_slicepitch);
-
-          dst_dev->ops->write_rect
-            (dst_dev->data,
-             tmp,
-             dst_buf->device_ptrs[dst_dev->dev_id].mem_ptr,
-             node->command.copy_image.dst_origin,
-             null_origin,
-             node->command.copy_image.region,
-             node->command.copy_image.dst_rowpitch,
-             node->command.copy_image.dst_slicepitch,
-             tmp_rowpitch,
-             tmp_slicepitch);
-          free (tmp);
-        }
-      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Copy Buffer Rect      ");
+      assert (dev->ops->write_image_rect);
+      dev->ops->write_image_rect (
+          dev->data,
+          event->mem_objs[1],
+          cmd->write_image.dst_mem_id,
+          NULL,
+          cmd->write_image.src_mem_id,
+          cmd->write_image.origin,
+          cmd->write_image.region,
+          cmd->write_image.src_row_pitch,
+          cmd->write_image.src_slice_pitch,
+          cmd->write_image.src_offset);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "CopyBufferToImage       ");
       break;
+
+    case CL_COMMAND_WRITE_IMAGE:
+        POCL_UPDATE_EVENT_RUNNING(event);
+        assert (dev->ops->write_image_rect);
+        dev->ops->write_image_rect (
+            dev->data,
+            event->mem_objs[0],
+            cmd->write_image.dst_mem_id,
+            cmd->write_image.src_host_ptr,
+            NULL,
+            cmd->write_image.origin,
+            cmd->write_image.region,
+            cmd->write_image.src_row_pitch,
+            cmd->write_image.src_slice_pitch,
+            cmd->write_image.src_offset);
+        POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Write Image           ");
+        break;
+
+    case CL_COMMAND_COPY_IMAGE:
+        POCL_UPDATE_EVENT_RUNNING(event);
+        assert (dev->ops->copy_image_rect);
+        dev->ops->copy_image_rect(
+              dev->data,
+              event->mem_objs[0],
+              event->mem_objs[1],
+              cmd->copy_image.src_mem_id,
+              cmd->copy_image.dst_mem_id,
+              cmd->copy_image.src_origin,
+              cmd->copy_image.dst_origin,
+              cmd->copy_image.region);
+        POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Copy Image            ");
+        break;
+
+    case CL_COMMAND_FILL_IMAGE:
+      POCL_UPDATE_EVENT_RUNNING (event);
+      assert (dev->ops->fill_image);
+      dev->ops->fill_image
+        (dev->data,
+         event->mem_objs[0],
+         cmd->fill_image.mem_id,
+         cmd->fill_image.origin,
+         cmd->fill_image.region,
+         cmd->fill_image.fill_pixel,
+         cmd->fill_image.pixel_size);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Fill Image            ");
+      POCL_MEM_FREE (cmd->fill_image.fill_pixel);
+      break;
+
+    case CL_COMMAND_MAP_IMAGE:
+      POCL_UPDATE_EVENT_RUNNING(event);
+      POCL_LOCK_OBJ (event->mem_objs[0]);
+      assert (dev->ops->map_image != NULL);
+      dev->ops->map_image (dev->data,
+                           cmd->map.mem_id,
+                           event->mem_objs[0],
+                           cmd->map.mapping);
+     (event->mem_objs[0])->map_count++;
+      POCL_UNLOCK_OBJ (event->mem_objs[0]);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Map Image             ");
+      break;
+
     case CL_COMMAND_UNMAP_MEM_OBJECT:
       POCL_UPDATE_EVENT_RUNNING(event);
-      /* TODO: should we ensure the device global region is updated from
-         the host memory? How does the specs define it,
-         can the host_ptr be assumed to point to the host and the
-         device accessible memory or just point there until the
-         kernel(s) get executed or similar? */
-      /* Assume the region is automatically up to date. */
-      POCL_LOCK_OBJ (node->command.unmap.memobj);
-      if (node->device->ops->unmap_mem != NULL)
-        node->device->ops->unmap_mem (node->device->data,
-                                      (node->command.unmap.mapping)->host_ptr,
-                                      (node->command.unmap.memobj)
-                                          ->device_ptrs[node->device->dev_id]
-                                          .mem_ptr,
-                                      (node->command.unmap.mapping)->offset,
-                                      (node->command.unmap.mapping)->size);
-      assert ((node->command.unmap.mapping)->unmap_requested > 0);
-      DL_DELETE((node->command.unmap.memobj)->mappings, 
-                node->command.unmap.mapping);
-      (node->command.unmap.memobj)->map_count--;
-      POCL_MEM_FREE (node->command.unmap.mapping);
-      POCL_UNLOCK_OBJ (node->command.unmap.memobj);
+      POCL_LOCK_OBJ (event->mem_objs[0]);
+      if (event->mem_objs[0]->is_image == CL_FALSE
+          || IS_IMAGE1D_BUFFER (event->mem_objs[0]))
+        {
+          assert (dev->ops->unmap_mem != NULL);
+          dev->ops->unmap_mem (dev->data,
+                               cmd->unmap.mem_id,
+                               event->mem_objs[0],
+                               cmd->unmap.mapping);
+        }
+      else
+        {
+          assert (dev->ops->unmap_image != NULL);
+          dev->ops->unmap_image (dev->data,
+                                 cmd->unmap.mem_id,
+                                 event->mem_objs[0],
+                                 cmd->unmap.mapping);
+        }
+      assert ((cmd->unmap.mapping)->unmap_requested > 0);
+      DL_DELETE((event->mem_objs[0])->mappings,
+                cmd->unmap.mapping);
+      (event->mem_objs[0])->map_count--;
+      POCL_MEM_FREE (cmd->unmap.mapping);
+      POCL_UNLOCK_OBJ (event->mem_objs[0]);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Unmap Mem obj         ");
       break;
+
     case CL_COMMAND_NDRANGE_KERNEL:
       POCL_UPDATE_EVENT_RUNNING(event);
       assert (event == node->event);
-      node->device->ops->run(node->command.run.data, node);
+      assert (dev->ops->run);
+      dev->ops->run (dev->data, node);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Enqueue NDRange       ");
       pocl_ndrange_node_cleanup(node);
       break;
+
     case CL_COMMAND_NATIVE_KERNEL:
       POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->run_native(node->command.native.data, node);
-      pocl_native_kernel_cleanup(node);
+      assert (dev->ops->run_native);
+      dev->ops->run_native (dev->data, node);
+      POCL_MEM_FREE (node->command.native.args);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Native Kernel         ");
       break;
-    case CL_COMMAND_FILL_IMAGE:
-      POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->fill_rect 
-        (node->command.fill_image.data, 
-         node->command.fill_image.device_ptr,
-         node->command.fill_image.buffer_origin,
-         node->command.fill_image.region,
-         node->command.fill_image.rowpitch, 
-         node->command.fill_image.slicepitch,
-         node->command.fill_image.fill_pixel,
-         node->command.fill_image.pixel_size);
-      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Fill Image            ");
-      free(node->command.fill_image.fill_pixel);
-      break;
-    case CL_COMMAND_FILL_BUFFER:
-      POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->memfill
-        (node->command.memfill.ptr,
-         node->command.memfill.size,
-         node->command.memfill.offset,
-         node->command.memfill.pattern,
-         node->command.memfill.pattern_size);
-      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Fill Buffer           ");
-      pocl_aligned_free(node->command.memfill.pattern);
-      break;
+
     case CL_COMMAND_MARKER:
       POCL_UPDATE_EVENT_RUNNING(event);
       POCL_UPDATE_EVENT_COMPLETE(event);
       break;
+
     case CL_COMMAND_BARRIER:
       POCL_UPDATE_EVENT_RUNNING(event);
       POCL_UPDATE_EVENT_COMPLETE(event);
       break;
+
     case CL_COMMAND_SVM_FREE:
       POCL_UPDATE_EVENT_RUNNING(event);
-      if (node->command.svm_free.pfn_free_func)
-        node->command.svm_free.pfn_free_func(
-           node->command.svm_free.queue,
-           node->command.svm_free.num_svm_pointers,
-           node->command.svm_free.svm_pointers,
-           node->command.svm_free.data);
+      if (cmd->svm_free.pfn_free_func)
+        cmd->svm_free.pfn_free_func(
+           cmd->svm_free.queue,
+           cmd->svm_free.num_svm_pointers,
+           cmd->svm_free.svm_pointers,
+           cmd->svm_free.data);
       else
-        for (i=0; i < node->command.svm_free.num_svm_pointers; i++)
-          node->device->ops->free_ptr(node->device,
-                                      node->command.svm_free.svm_pointers[i]);
+        for (i = 0; i < cmd->svm_free.num_svm_pointers; i++)
+          dev->ops->svm_free (dev, cmd->svm_free.svm_pointers[i]);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "SVM Free              ");
       break;
+
     case CL_COMMAND_SVM_MAP:
       POCL_UPDATE_EVENT_RUNNING(event);
-      if (DEVICE_MMAP_IS_NOP(node->device))
+      if (DEVICE_MMAP_IS_NOP (dev))
         ; // no-op
       else
-        node->device->ops->map_mem
-          (node->device->data, node->command.svm_map.svm_ptr,
-           0, node->command.svm_map.size, NULL);
+        {
+          assert (dev->ops->svm_map);
+          dev->ops->svm_map (dev, cmd->svm_map.svm_ptr);
+        }
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "SVM Map              ");
       break;
+
     case CL_COMMAND_SVM_UNMAP:
       POCL_UPDATE_EVENT_RUNNING(event);
-      if (DEVICE_MMAP_IS_NOP(node->device))
+      if (DEVICE_MMAP_IS_NOP (dev))
         ; // no-op
       else
-        node->device->ops->unmap_mem
-          (node->device->data, NULL,
-           node->command.svm_unmap.svm_ptr, 0, 0);
+        {
+          assert (dev->ops->svm_unmap);
+          dev->ops->svm_unmap (dev, cmd->svm_unmap.svm_ptr);
+        }
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "SVM Unmap             ");
       break;
+
     case CL_COMMAND_SVM_MEMCPY:
       POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->copy(NULL,
-                              node->command.svm_memcpy.src, 0,
-                              node->command.svm_memcpy.dst, 0,
-                              node->command.svm_memcpy.size);
+      assert (dev->ops->svm_copy);
+      dev->ops->svm_copy (dev,
+                          cmd->svm_memcpy.dst,
+                          cmd->svm_memcpy.src,
+                          cmd->svm_memcpy.size);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "SVM Memcpy            ");
       break;
+
     case CL_COMMAND_SVM_MEMFILL:
       POCL_UPDATE_EVENT_RUNNING(event);
-      node->device->ops->memfill(
-                                 node->command.memfill.ptr,
-                                 node->command.memfill.size, 0,
-                                 node->command.memfill.pattern,
-                                 node->command.memfill.pattern_size);
+      assert (dev->ops->svm_fill);
+      dev->ops->svm_fill (dev,
+                          cmd->svm_fill.svm_ptr,
+                          cmd->svm_fill.size,
+                          cmd->svm_fill.pattern,
+                          cmd->svm_fill.pattern_size);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "SVM MemFill           ");
-      pocl_aligned_free(node->command.memfill.pattern);
+      pocl_aligned_free (cmd->svm_fill.pattern);
       break;
+
     default:
       POCL_ABORT_UNIMPLEMENTED("");
       break;
@@ -705,6 +758,7 @@ pocl_exec_command (_cl_command_node * volatile node)
   pocl_mem_manager_free_command (node);
 }
 
+/* call with brc_event UNLOCKED. */
 void
 pocl_broadcast (cl_event brc_event)
 {
@@ -773,21 +827,6 @@ fill_dev_sampler_t (dev_sampler_t *ds, struct pocl_argument *parg)
   }
 }
 
-void*
-pocl_memalign_alloc(size_t align_width, size_t size)
-{
-  void *ptr;
-  int status;
-
-#ifndef POCL_ANDROID
-  status = posix_memalign(&ptr, align_width, size);
-  return ((status == 0)? ptr: (void*)NULL);
-#else
-  ptr = memalign(align_width, size);
-  return ptr;
-#endif
-}
-
 /* CPU driver stuff */
 typedef struct pocl_dlhandle_cache_item pocl_dlhandle_cache_item;
 struct pocl_dlhandle_cache_item
@@ -932,8 +971,8 @@ pocl_check_dlhandle_cache (_cl_command_node *cmd, unsigned initial_refcount)
           POCL_ABORT ("Final linking of kernel %s failed.\n", k->name);
         }
 #else
-      POCL_ABORT("pocl built without online compiler support "
-                 "cannot compile LLVM IRs to machine code\n");
+      POCL_ABORT("pocl built without online compiler support"
+                 " cannot compile LLVM IRs to machine code\n");
 #endif
     }
   else
@@ -993,6 +1032,14 @@ pocl_check_dlhandle_cache (_cl_command_node *cmd, unsigned initial_refcount)
                   "_pocl_launcher_%s_workgroup", k->name);
         ci->wg = (pocl_workgroup)lt_dlsym (ci->dlhandle, workgroup_string);
         dl_error = lt_dlerror ();
+        if (ci->wg == NULL)
+          {
+            // Older osx dyld APIs need the name without the underscore
+            snprintf (workgroup_string, WORKGROUP_STRING_LENGTH,
+                      "pocl_launcher_%s_workgroup", k->name);
+            ci->wg = (pocl_workgroup)lt_dlsym (ci->dlhandle, workgroup_string);
+            dl_error = lt_dlerror ();
+          }
 
         if (ci->wg != NULL && dl_error == NULL)
           {
@@ -1020,7 +1067,7 @@ pocl_check_dlhandle_cache (_cl_command_node *cmd, unsigned initial_refcount)
 #define MIN_MAX_MEM_ALLOC_SIZE (128*1024*1024)
 
 /* accounting object for the main memory */
-static pocl_global_mem_t system_memory;
+static pocl_global_mem_t system_memory = {POCL_LOCK_INITIALIZER, 0, 0, 0};
 
 void
 pocl_setup_device_for_system_memory(cl_device_id device)
@@ -1056,8 +1103,8 @@ pocl_setup_device_for_system_memory(cl_device_id device)
       if (device->global_mem_size > limited_memory)
         device->global_mem_size = limited_memory;
       else
-        POCL_MSG_WARN ("requested POCL_MEMORY_LIMIT %i GBs is larger than "
-                       "physical memory size (%zu) GBs, ignoring\n",
+        POCL_MSG_WARN ("requested POCL_MEMORY_LIMIT %i GBs is larger than"
+                       " physical memory size (%zu) GBs, ignoring\n",
                        limit_memory_gb, (device->global_mem_size >> 30));
     }
 
@@ -1155,7 +1202,7 @@ pocl_set_buffer_image_limits(cl_device_id device)
 }
 
 void*
-pocl_memalign_alloc_global_mem(cl_device_id device, size_t align, size_t size)
+pocl_aligned_malloc_global_mem(cl_device_id device, size_t align, size_t size)
 {
   pocl_global_mem_t *mem = device->global_memory;
   void *retval = NULL;
@@ -1164,7 +1211,7 @@ pocl_memalign_alloc_global_mem(cl_device_id device, size_t align, size_t size)
   if ((mem->total_alloc_limit - mem->currently_allocated) < size)
     goto ERROR;
 
-  retval = pocl_memalign_alloc (align, size);
+  retval = pocl_aligned_malloc (align, size);
   if (!retval)
     goto ERROR;
 
@@ -1192,6 +1239,7 @@ pocl_free_global_mem(cl_device_id device, void* ptr, size_t size)
   POCL_MEM_FREE(ptr);
 }
 
+
 void
 pocl_print_system_memory_stats()
 {
