@@ -26,12 +26,13 @@
 #define POCL_COMMON_H
 
 #include "pocl_cl.h"
-#include "dev_image.h"
 #include "utlist.h"
 
+#define __CBUILD__
+#include "pocl_image_types.h"
+#undef __CBUILD__
+
 #define XSETUP_DEVICE_CL_VERSION(A, B)             \
-  dev->cl_version_major = A;                      \
-  dev->cl_version_minor = B;                      \
   dev->cl_version_int = (A * 100) + (B * 10);     \
   dev->cl_version_std = "CL" # A "." # B;         \
   dev->version = "OpenCL " # A "." # B " pocl";
@@ -63,6 +64,8 @@ extern "C" {
 #pragma GCC visibility push(hidden)
 #endif
 
+void pocl_init_cpu_device_infos (cl_device_id dev);
+
 char *llvm_codegen (const char *tmpdir, cl_kernel kernel, cl_device_id device,
                     size_t local_x, size_t local_y, size_t local_z);
 
@@ -71,9 +74,7 @@ void fill_dev_image_t (dev_image_t* di, struct pocl_argument* parg,
 
 void fill_dev_sampler_t (dev_sampler_t *ds, struct pocl_argument *parg);
 
-void* pocl_memalign_alloc(size_t align_width, size_t size);
-
-void pocl_copy_mem_object (cl_device_id dest_dev, cl_mem dest, 
+void pocl_copy_mem_object (cl_device_id dest_dev, cl_mem dest,
                            size_t dest_offset,
                            cl_device_id source_dev, cl_mem source,
                            size_t source_offset, size_t cb);
@@ -86,7 +87,6 @@ void pocl_scheduler (_cl_command_node * volatile * ready_list,
 void pocl_exec_command (_cl_command_node * volatile node);
 
 void pocl_ndrange_node_cleanup(_cl_command_node *node);
-void pocl_native_kernel_cleanup(_cl_command_node *node);
 void pocl_mem_objs_cleanup (cl_event event);
 
 void pocl_broadcast (cl_event event);
@@ -102,7 +102,7 @@ void pocl_setup_device_for_system_memory(cl_device_id device);
 
 void pocl_set_buffer_image_limits(cl_device_id device);
 
-void* pocl_memalign_alloc_global_mem(cl_device_id device, size_t align, size_t size);
+void* pocl_aligned_malloc_global_mem(cl_device_id device, size_t align, size_t size);
 
 void pocl_free_global_mem(cl_device_id device, void *ptr, size_t size);
 
