@@ -45,7 +45,9 @@ extern unsigned int pocl_num_devices;
  * The devices are shared across contexts, thus must implement resource
  * management internally also across multiple contexts.
  */
-void pocl_init_devices();
+cl_int pocl_init_devices();
+
+cl_int pocl_uninit_devices ();
 
 /**
  * \brief Get the count of devices for a specific type
@@ -71,6 +73,13 @@ unsigned int pocl_get_devices(cl_device_type device_type, struct _cl_device_id *
  */
 int pocl_device_get_env_count(const char *dev_type);
 
+
+/**
+ * \brief Unique global memory id for devices with distinct memory from the system memory
+ * \return Unique global mem id, id > 0. Zero is reserved for shared system memory
+ */
+int pocl_get_unique_global_mem_id();
+
 #ifdef __GNUC__
 #pragma GCC visibility pop
 #endif
