@@ -199,22 +199,20 @@ pocl_hsa_init_device_ops(struct pocl_device_ops *ops)
 
   /* TODO: more descriptive name from HSA probing the device */
   ops->device_name = "hsa";
-  ops->init_device_infos = pocl_hsa_init_device_infos;
   ops->probe = pocl_hsa_probe;
   ops->uninit = pocl_hsa_uninit;
   ops->reinit = NULL;
   ops->init = pocl_hsa_init;
   ops->alloc_mem_obj = pocl_hsa_alloc_mem_obj;
   ops->free = pocl_hsa_free;
-  //ops->run = pocl_hsa_run;
-  ops->read = pocl_basic_read;
-  ops->read_rect = pocl_basic_read_rect;
-  ops->write = pocl_basic_write;
-  ops->write_rect = pocl_basic_write_rect;
+  ops->run = NULL;
   ops->copy = pocl_hsa_copy;
-  ops->copy_rect = pocl_basic_copy_rect;
   ops->get_timer_value = pocl_hsa_get_timer_value;
 
+  ops->svm_free = pocl_hsa_svm_free;
+  ops->svm_alloc = pocl_hsa_svm_alloc;
+  ops->svm_copy = pocl_hsa_svm_copy;
+
   // new driver api (out-of-order)
   ops->submit = pocl_hsa_submit;
   ops->join = pocl_hsa_join;
@@ -226,11 +224,8 @@ pocl_hsa_init_device_ops(struct pocl_device_ops *ops)
   ops->update_event = pocl_hsa_update_event;
   ops->free_event_data = pocl_hsa_free_event_data;
   ops->init_target_machine = NULL;
-  ops->run = NULL;
   ops->wait_event = pocl_hsa_wait_event;
   ops->update_event = pocl_hsa_update_event;
-  ops->free_event_data = NULL;
-
   ops->build_hash = pocl_hsa_build_hash;
 }
 
@@ -412,6 +407,7 @@ get_hsa_device_features(char* dev_name, struct _cl_device_id* dev)
     {
       if (strcmp(dev_name, supported_hsa_devices[i].long_name) == 0)
         {
+          dev->device_side_printf = 0;
           COPY_ATTR (llvm_cpu);
           COPY_ATTR (llvm_target_triplet);
           COPY_ATTR (has_64bit_long);
@@ -445,10 +441,40 @@ get_hsa_device_features(char* dev_name, struct _cl_device_id* dev)
     }
 }
 
-void
-pocl_hsa_init_device_infos(unsigned j, struct _cl_device_id* dev)
+unsigned int
+pocl_hsa_probe (struct pocl_device_ops *ops)
 {
-  pocl_basic_init_device_infos (j, dev);
+  int env_count = pocl_device_get_env_count (ops->device_name);
+
+  POCL_MSG_PRINT_INFO ("pocl-hsa: found %d env devices with %s.\n", env_count,
+                       ops->device_name);
+
+  /* No hsa env specified, the user did not request for HSA agents. */
+  if (env_count <= 0)
+    return 0;
+
+  HSA_CHECK (hsa_init ());
+
+  HSA_CHECK (hsa_iterate_agents (pocl_hsa_get_agents_callback, NULL));
+
+  POCL_MSG_PRINT_INFO ("pocl-hsa: found %d agents.\n", found_hsa_agents);
+
+  return (int)found_hsa_agents;
+}
+
+static void
+hsa_queue_callback (hsa_status_t status, hsa_queue_t *q, void *data)
+{
+  HSA_CHECK (status);
+}
+
+/* driver pthread prototype */
+void *pocl_hsa_driver_pthread (void *cldev);
+
+cl_int
+pocl_hsa_init (unsigned j, cl_device_id dev, const char *parameters)
+{
+  pocl_init_cpu_device_infos (dev);
 
   SETUP_DEVICE_CL_VERSION(HSA_DEVICE_CL_VERSION_MAJOR,
                           HSA_DEVICE_CL_VERSION_MINOR)
@@ -534,32 +560,29 @@ pocl_hsa_init_device_infos(unsigned j, struct _cl_device_id* dev)
   if (dev->image_support == CL_TRUE)
     {
       hsa_dim3_t image_size;
-      HSA_CHECK(hsa_agent_get_info (agent,
-				    HSA_EXT_AGENT_INFO_IMAGE_1D_MAX_ELEMENTS,
-				    &image_size));
+      HSA_CHECK (hsa_agent_get_info (
+          agent, HSA_EXT_AGENT_INFO_IMAGE_1D_MAX_ELEMENTS, &image_size));
       dev->image_max_buffer_size = image_size.x;
-      HSA_CHECK(hsa_agent_get_info (agent,
-				    HSA_EXT_AGENT_INFO_IMAGE_2D_MAX_ELEMENTS,
-				    &image_size));
+      HSA_CHECK (hsa_agent_get_info (
+          agent, HSA_EXT_AGENT_INFO_IMAGE_2D_MAX_ELEMENTS, &image_size));
       dev->image2d_max_height = image_size.x;
       dev->image2d_max_width = image_size.y;
-      HSA_CHECK(hsa_agent_get_info (agent,
-				    HSA_EXT_AGENT_INFO_IMAGE_3D_MAX_ELEMENTS,
-				    &image_size));
+      HSA_CHECK (hsa_agent_get_info (
+          agent, HSA_EXT_AGENT_INFO_IMAGE_3D_MAX_ELEMENTS, &image_size));
       dev->image3d_max_height = image_size.x;
       dev->image3d_max_width = image_size.y;
       dev->image3d_max_depth = image_size.z;
       // is this directly the product of the dimensions?
       //stat = hsa_agent_get_info(agent, ??, &dev->image_max_array_size);
-      HSA_CHECK(hsa_agent_get_info
-		(agent, HSA_EXT_AGENT_INFO_MAX_IMAGE_RD_HANDLES,
-		 &dev->max_read_image_args));
-      HSA_CHECK(hsa_agent_get_info
-		(agent, HSA_EXT_AGENT_INFO_MAX_IMAGE_RORW_HANDLES,
-		 &dev->max_read_write_image_args));
+      HSA_CHECK (hsa_agent_get_info (agent,
+                                     HSA_EXT_AGENT_INFO_MAX_IMAGE_RD_HANDLES,
+                                     &dev->max_read_image_args));
+      HSA_CHECK (hsa_agent_get_info (agent,
+                                     HSA_EXT_AGENT_INFO_MAX_IMAGE_RORW_HANDLES,
+                                     &dev->max_read_write_image_args));
       dev->max_write_image_args = dev->max_read_write_image_args;
-      HSA_CHECK(hsa_agent_get_info
-		(agent, HSA_EXT_AGENT_INFO_MAX_SAMPLER_HANDLERS, &dev->max_samplers));
+      HSA_CHECK (hsa_agent_get_info (
+          agent, HSA_EXT_AGENT_INFO_MAX_SAMPLER_HANDLERS, &dev->max_samplers));
     }
 
   dev->should_allocate_svm = 1;
@@ -579,51 +602,17 @@ pocl_hsa_init_device_infos(unsigned j, struct _cl_device_id* dev)
                                | CL_QUEUE_PROFILING_ENABLE;
   dev->on_host_queue_props = CL_QUEUE_PROFILING_ENABLE;
 
-}
-
-unsigned int
-pocl_hsa_probe(struct pocl_device_ops *ops)
-{
-  int env_count = pocl_device_get_env_count(ops->device_name);
-
-  POCL_MSG_PRINT_INFO("pocl-hsa: found %d env devices with %s.\n",
-                      env_count, ops->device_name);
-
-  /* No hsa env specified, the user did not request for HSA agents. */
-  if (env_count <= 0)
-    return 0;
-
-  HSA_CHECK(hsa_init());
-
-  HSA_CHECK(hsa_iterate_agents(pocl_hsa_get_agents_callback, NULL));
-
-  POCL_MSG_PRINT_INFO("pocl-hsa: found %d agents.\n", found_hsa_agents);
-
-  return (int)found_hsa_agents;
-}
-
-static void
-hsa_queue_callback(hsa_status_t status, hsa_queue_t *q, void* data) {
-  HSA_CHECK(status);
-}
-
-/* driver pthread prototype */
-void * pocl_hsa_driver_pthread (void *cldev);
-
-cl_int
-pocl_hsa_init (unsigned j, cl_device_id device, const char* parameters)
-{
   pocl_hsa_device_data_t *d;
 
-  device->global_mem_id = 0;
+  dev->global_mem_id = 0;
 
   d = (pocl_hsa_device_data_t *) calloc (1, sizeof(pocl_hsa_device_data_t));
 
   POCL_INIT_LOCK (d->pocl_hsa_compilation_lock);
 
-  intptr_t agent_index = (intptr_t)device->data;
+  intptr_t agent_index = (intptr_t)dev->data;
   d->agent.handle = hsa_agents[agent_index].handle;
-  device->data = d;
+  dev->data = d;
 
   HSA_CHECK(hsa_agent_iterate_regions (d->agent,
                                        setup_agent_memory_regions_callback,
@@ -646,7 +635,7 @@ pocl_hsa_init (unsigned j, cl_device_id device, const char* parameters)
   size_t sizearg;
   HSA_CHECK(hsa_region_get_info(d->global_region,
 				HSA_REGION_INFO_ALLOC_MAX_SIZE, &sizearg));
-  device->max_mem_alloc_size = sizearg;
+  dev->max_mem_alloc_size = sizearg;
 
   /* For some reason, the global region size returned is 128 Terabytes...
    * for now, use the max alloc size, it seems to be a much more reasonable
@@ -656,27 +645,25 @@ pocl_hsa_init (unsigned j, cl_device_id device, const char* parameters)
    */
   HSA_CHECK(hsa_region_get_info(d->global_region,
                                HSA_REGION_INFO_SIZE, &sizearg));
-  device->global_mem_size = sizearg;
-	if (device->global_mem_size > 16*1024*1024*(uint64_t)1024)
-		device->global_mem_size = device->max_mem_alloc_size;
+  dev->global_mem_size = sizearg;
+  if (dev->global_mem_size > 16 * 1024 * 1024 * (uint64_t)1024)
+    dev->global_mem_size = dev->max_mem_alloc_size;
 
-  pocl_setup_device_for_system_memory(device);
+  pocl_setup_device_for_system_memory (dev);
 
   HSA_CHECK(hsa_region_get_info(d->group_region, HSA_REGION_INFO_SIZE,
                                 &sizearg));
-  device->local_mem_size = sizearg;
+  dev->local_mem_size = sizearg;
 
   HSA_CHECK(hsa_region_get_info(d->global_region,
 				HSA_REGION_INFO_RUNTIME_ALLOC_ALIGNMENT,
                                 &sizearg));
-  device->mem_base_addr_align = sizearg * 8;
+  dev->mem_base_addr_align = sizearg * 8;
 
   HSA_CHECK(hsa_agent_get_info(d->agent, HSA_AGENT_INFO_PROFILE,
                                &d->agent_profile));
-  device->profile = (
-      (d->agent_profile == HSA_PROFILE_FULL) ?
-      "FULL_PROFILE" : "EMBEDDED_PROFILE");
-
+  dev->profile = ((d->agent_profile == HSA_PROFILE_FULL) ? "FULL_PROFILE"
+                                                         : "EMBEDDED_PROFILE");
 
   uint64_t hsa_freq;
   HSA_CHECK(hsa_system_get_info(HSA_SYSTEM_INFO_TIMESTAMP_FREQUENCY,
@@ -687,7 +674,7 @@ pocl_hsa_init (unsigned j, cl_device_id device, const char* parameters)
   POCL_MSG_PRINT_INFO("HSA timeout: %" PRIu64 "\n", d->timeout);
   POCL_MSG_PRINT_INFO("HSA timestamp unit: %g\n", d->timestamp_unit);
 
-  device->profiling_timer_resolution = (size_t)(d->timestamp_unit) || 1;
+  dev->profiling_timer_resolution = (size_t) (d->timestamp_unit) || 1;
 
   /* TODO proper setup */
   d->hw_schedulers = 3;
@@ -705,8 +692,8 @@ pocl_hsa_init (unsigned j, cl_device_id device, const char* parameters)
   PTHREAD_CHECK(pthread_mutex_init(&d->list_mutex, &mattr));
 
   d->exit_driver_thread = 0;
-  PTHREAD_CHECK(pthread_create(&d->driver_pthread_id, NULL,
-                 &pocl_hsa_driver_pthread, device));
+  PTHREAD_CHECK (pthread_create (&d->driver_pthread_id, NULL,
+                                 &pocl_hsa_driver_pthread, dev));
   return CL_SUCCESS;
 }
 
@@ -787,12 +774,20 @@ pocl_hsa_free (cl_device_id device, cl_mem memobj)
 }
 
 void
-pocl_hsa_copy (void *data, const void *src_ptr, size_t src_offset,
-               void *__restrict__ dst_ptr, size_t dst_offset, size_t cb)
+pocl_hsa_copy (void *data,
+               pocl_mem_identifier * dst_mem_id,
+               cl_mem dst_buf,
+               pocl_mem_identifier * src_mem_id,
+               cl_mem src_buf,
+               size_t dst_offset,
+               size_t src_offset,
+               size_t size)
 {
+  void *__restrict__ dst_ptr = dst_mem_id->mem_ptr;
+  void *__restrict__ src_ptr = src_mem_id->mem_ptr;
   assert(src_offset == 0);
   assert(dst_offset == 0);
-  HSA_CHECK(hsa_memory_copy(dst_ptr, src_ptr, cb));
+  HSA_CHECK (hsa_memory_copy (dst_ptr, src_ptr, size));
 }
 
 cl_int
@@ -1241,7 +1236,7 @@ pocl_hsa_join (cl_device_id device, cl_command_queue cq)
   cl_event event = cq->last_event.event;
   assert(event);
   POCL_LOCK_OBJ (event);
-  ++event->pocl_refcount;
+  POCL_RETAIN_OBJECT_UNLOCKED (event);
   POCL_UNLOCK_OBJ (cq);
 
   POCL_MSG_PRINT_INFO("pocl-hsa: device->join on event %u\n", event->id);
@@ -1250,9 +1245,7 @@ pocl_hsa_join (cl_device_id device, cl_command_queue cq)
     {
       POCL_MSG_PRINT_INFO("pocl-hsa: device->join: last event (%u) in queue"
                           " exists, but is complete\n", event->id);
-      --event->pocl_refcount;
-      POCL_UNLOCK_OBJ (event);
-      return;
+      goto RETURN;
     }
 
   while (event->status > CL_COMPLETE)
@@ -1263,9 +1256,11 @@ pocl_hsa_join (cl_device_id device, cl_command_queue cq)
   POCL_MSG_PRINT_INFO("pocl-hsa: device->join on event %u finished"
                       " with status: %i\n", event->id, event->status);
 
+RETURN:
   assert (event->status <= CL_COMPLETE);
-  --event->pocl_refcount;
   POCL_UNLOCK_OBJ (event);
+
+  POname (clReleaseEvent) (event);
 }
 
 void
@@ -1478,6 +1473,7 @@ pocl_hsa_ndrange_event_finished (pocl_hsa_device_data_t *d, size_t i)
 
   assert(i < dd->running_list_size);
   cl_event event = dd->running_events[i];
+  _cl_command_node *node = event->command;
 
   POCL_LOCK_OBJ (event);
   pocl_hsa_event_data_t *event_data = (pocl_hsa_event_data_t *)event->data;
@@ -1501,11 +1497,12 @@ pocl_hsa_ndrange_event_finished (pocl_hsa_device_data_t *d, size_t i)
 
   hsa_memory_free(event_data->actual_kernargs);
 
-  pocl_mem_manager_free_command (event->command);
-
   POCL_UNLOCK_OBJ (event);
+
   POCL_UPDATE_EVENT_COMPLETE (event);
 
+  pocl_ndrange_node_cleanup (node);
+  pocl_mem_manager_free_command (node);
 }
 
 static void
@@ -1753,3 +1750,38 @@ void pocl_hsa_free_event_data (cl_event event)
   free(event->data);
   event->data = NULL;
 }
+
+/****** SVM callbacks *****/
+
+void
+pocl_hsa_svm_free (cl_device_id dev, void *svm_ptr)
+{
+  /* TODO we should somehow figure out the size argument
+   * and call pocl_free_global_mem */
+  HSA_CHECK (hsa_memory_free (svm_ptr));
+}
+
+void *
+pocl_hsa_svm_alloc (cl_device_id dev, cl_svm_mem_flags flags, size_t size)
+{
+  POCL_RETURN_ERROR_ON (((flags & CL_MEM_SVM_ATOMICS)
+                         && ((dev->svm_caps & CL_DEVICE_SVM_ATOMICS) == 0)),
+                        NULL, "This device doesn't have SVM Atomics");
+
+  POCL_RETURN_ERROR_ON (
+      ((flags & CL_MEM_SVM_FINE_GRAIN_BUFFER)
+       && ((dev->svm_caps & CL_DEVICE_SVM_FINE_GRAIN_BUFFER) == 0)),
+      NULL, "This device doesn't have SVM Atomics");
+
+  pocl_hsa_device_data_t *d = (pocl_hsa_device_data_t *)dev->data;
+  void *b = NULL;
+  HSA_CHECK (hsa_memory_allocate (d->global_region, size, &b));
+  return b;
+}
+
+void
+pocl_hsa_svm_copy (cl_device_id dev, void *__restrict__ dst,
+                   const void *__restrict__ src, size_t size)
+{
+  HSA_CHECK (hsa_memory_copy (dst, src, size));
+}
