@@ -76,7 +76,6 @@ pocl_ttasim_init_device_ops(struct pocl_device_ops *ops)
   ops->device_name = "ttasim";
 
   ops->probe = pocl_ttasim_probe;
-  ops->init_device_infos = pocl_ttasim_init_device_infos;
   ops->uninit = pocl_ttasim_uninit;
   ops->reinit = NULL;
   ops->init = pocl_ttasim_init;
@@ -90,6 +89,7 @@ pocl_ttasim_init_device_ops(struct pocl_device_ops *ops)
   ops->copy = pocl_tce_copy;
   ops->copy_rect = pocl_tce_copy_rect;
   ops->map_mem = pocl_tce_map_mem;
+  ops->unmap_mem = pocl_tce_unmap_mem;
   ops->run = pocl_tce_run;
   ops->get_timer_value = pocl_ttasim_get_timer_value;
   ops->init_build = pocl_tce_init_build;
@@ -105,78 +105,6 @@ pocl_ttasim_init_device_ops(struct pocl_device_ops *ops)
 }
 
 
-void
-pocl_ttasim_init_device_infos(unsigned j, struct _cl_device_id* dev)
-{
-  dev->type = CL_DEVICE_TYPE_GPU;
-  dev->max_compute_units = 1;
-  dev->max_work_item_dimensions = 3;
-
-  int max_wg
-      = pocl_get_int_option("POCL_MAX_WORK_GROUP_SIZE", DEFAULT_WG_SIZE);
-  assert(max_wg > 0);
-  max_wg = std::min(max_wg, DEFAULT_WG_SIZE);
-  if (max_wg < 0)
-    max_wg = DEFAULT_WG_SIZE;
-
-  dev->max_work_item_sizes[0] = dev->max_work_item_sizes[1]
-      = dev->max_work_item_sizes[2] = dev->max_work_group_size = max_wg;
-
-  dev->preferred_wg_size_multiple = 8;
-  dev->preferred_vector_width_char = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_CHAR;
-  dev->preferred_vector_width_short = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_SHORT;
-  dev->preferred_vector_width_int = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_INT;
-  dev->preferred_vector_width_long = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_LONG;
-  dev->preferred_vector_width_float = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_FLOAT;
-  dev->preferred_vector_width_double = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_DOUBLE;
-  dev->preferred_vector_width_half = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_HALF;
-  /* TODO: figure out what the difference between preferred and native widths are. */
-  dev->native_vector_width_char = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_CHAR;
-  dev->native_vector_width_short = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_SHORT;
-  dev->native_vector_width_int = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_INT;
-  dev->native_vector_width_long = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_LONG;
-  dev->native_vector_width_float = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_FLOAT;
-  dev->native_vector_width_double = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_DOUBLE;
-  dev->native_vector_width_half = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_HALF;
-  dev->max_clock_frequency = 100;
-  dev->image_support = CL_FALSE;
-  dev->single_fp_config = CL_FP_ROUND_TO_NEAREST | CL_FP_INF_NAN;
-  dev->double_fp_config = CL_FP_ROUND_TO_NEAREST | CL_FP_INF_NAN;
-  dev->global_mem_cache_type = CL_NONE;
-  dev->local_mem_type = CL_GLOBAL;
-  dev->error_correction_support = CL_FALSE;
-  dev->host_unified_memory = CL_FALSE;
-  dev->endian_little = CL_FALSE;
-  dev->available = CL_TRUE;
-  dev->compiler_available = CL_TRUE;
-  dev->spmd = CL_FALSE;
-  dev->workgroup_pass = CL_TRUE;
-  dev->execution_capabilities = CL_EXEC_KERNEL;
-  dev->queue_properties = CL_QUEUE_PROFILING_ENABLE;
-  dev->vendor = "TTA-Based Co-design Environment";
-  dev->profile = "EMBEDDED_PROFILE";
-  dev->extensions = TCE_DEVICE_EXTENSIONS;
-  dev->llvm_target_triplet = "tce-tut-llvm";
-  dev->has_64bit_long = 1;
-  dev->autolocals_to_args = 1;
-
-  dev->global_as_id = 3;
-  dev->local_as_id = 4;
-  dev->constant_as_id = 5;
-
-  SETUP_DEVICE_CL_VERSION(TCE_DEVICE_CL_VERSION_MAJOR, TCE_DEVICE_CL_VERSION_MINOR);
-
-  dev->parent_device = NULL;
-  // ttasim does not support partitioning
-  dev->max_sub_devices = 1;
-  dev->num_partition_properties = 1;
-  dev->partition_properties = (cl_device_partition_property*)calloc(dev->num_partition_properties,
-    sizeof(cl_device_partition_property));
-  dev->num_partition_types = 0;
-  dev->partition_type = NULL;
-
-}
-
 unsigned int
 pocl_ttasim_probe(struct pocl_device_ops *ops)
 {
@@ -208,6 +136,19 @@ public:
     Application::setSignalHandler(SIGINT, *ctrlcHandler);
 
     setMachine(simulator.machine());
+    if (machine_->isLittleEndian()) {
+      dev->endian_little = CL_TRUE;
+      dev->llvm_target_triplet = "tcele-tut-llvm";
+    } else {
+      dev->endian_little = CL_FALSE;
+      dev->llvm_target_triplet = "tce-tut-llvm";
+    }
+
+#if defined(WORDS_BIGENDIAN) && WORDS_BIGENDIAN == 1
+    needsByteSwap = ((dev->endian_little == CL_TRUE) ? true : false);
+#else
+    needsByteSwap = ((dev->endian_little == CL_TRUE) ? false : true);
+#endif
 
     initMemoryManagement(simulator.machine());
 
@@ -230,7 +171,15 @@ public:
       unsigned char val = ((char*)host_ptr)[i];
       globalMem->write (dest_addr + i, (Memory::MAU)(val));
     }
+  }
 
+  virtual void copyDeviceToDevice(uint32_t src_addr, uint32_t dest_addr, size_t count) {
+    MemorySystem &mems = simulator.memorySystem();
+    MemorySystem::MemoryPtr globalMem = mems.memory (*global_as);
+    for (std::size_t i = 0; i < count; ++i) {
+      unsigned char val =  globalMem->read (src_addr + i);
+      globalMem->write (dest_addr + i, (Memory::MAU)(val));
+    }
   }
 
   virtual void copyDeviceToHost(uint32_t src_addr, const void *host_ptr, 
@@ -285,12 +234,12 @@ public:
 
     std::ofstream out(fname.c_str());
 
-    out << "#include <lwpr.h>" << std::endl;
+    out << " <lwpr.h>" << std::endl;
     out << "#include <pocl_device.h>" << std::endl << std::endl;
 
-    out << "#define __local__ __attribute__((address_space(0)))" << std::endl;
-    out << "#define __global__ __attribute__((address_space(3)))" << std::endl;
-    out << "#define __constant__ __attribute__((address_space(3)))" << std::endl << std::endl;
+    out << "#define __local__ __attribute__((address_space(" <<  TTA_ASID_LOCAL<< ")))" << std::endl;
+    out << "#define __global__ __attribute__((address_space(" << TTA_ASID_GLOBAL << ")))" << std::endl;
+    out << "#define __constant__ __attribute__((address_space( " << TTA_ASID_CONSTANT << " )))" << std::endl << std::endl;
     out << "typedef volatile __global__ __kernel_exec_cmd kernel_exec_cmd;" << std::endl;
 
     /* Need to byteswap back as we are writing C code. */
@@ -543,13 +492,86 @@ private:
 
 
 cl_int
-pocl_ttasim_init (unsigned j, cl_device_id device, const char* parameters)
+pocl_ttasim_init (unsigned j, cl_device_id dev, const char* parameters)
 {
   if (parameters == NULL)
     POCL_ABORT("The tta device requires the adf file as a device parameter.\n"
                "Set it with POCL_TTASIMn_PARAMETERS=\"path/to/themachine.adf\".\n");
 
-  new TTASimDevice(device, parameters);
+  dev->type = CL_DEVICE_TYPE_GPU;
+  dev->max_compute_units = 1;
+  dev->max_work_item_dimensions = 3;
+  dev->device_side_printf = 0;
+
+  int max_wg
+      = pocl_get_int_option ("POCL_MAX_WORK_GROUP_SIZE", DEFAULT_WG_SIZE);
+  assert (max_wg > 0);
+  max_wg = std::min (max_wg, DEFAULT_WG_SIZE);
+  if (max_wg < 0)
+    max_wg = DEFAULT_WG_SIZE;
+
+  dev->max_work_item_sizes[0] = dev->max_work_item_sizes[1]
+      = dev->max_work_item_sizes[2] = dev->max_work_group_size = max_wg;
+
+  dev->preferred_wg_size_multiple = 8;
+  dev->preferred_vector_width_char = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_CHAR;
+  dev->preferred_vector_width_short
+      = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_SHORT;
+  dev->preferred_vector_width_int = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_INT;
+  dev->preferred_vector_width_long = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_LONG;
+  dev->preferred_vector_width_float
+      = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_FLOAT;
+  dev->preferred_vector_width_double
+      = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_DOUBLE;
+  dev->preferred_vector_width_half = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_HALF;
+  /* TODO: figure out what the difference between preferred and native widths
+   * are. */
+  dev->native_vector_width_char = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_CHAR;
+  dev->native_vector_width_short = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_SHORT;
+  dev->native_vector_width_int = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_INT;
+  dev->native_vector_width_long = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_LONG;
+  dev->native_vector_width_float = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_FLOAT;
+  dev->native_vector_width_double = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_DOUBLE;
+  dev->native_vector_width_half = POCL_DEVICES_PREFERRED_VECTOR_WIDTH_HALF;
+  dev->max_clock_frequency = 100;
+  dev->image_support = CL_FALSE;
+  dev->single_fp_config = CL_FP_ROUND_TO_NEAREST | CL_FP_INF_NAN;
+  dev->double_fp_config = CL_FP_ROUND_TO_NEAREST | CL_FP_INF_NAN;
+  dev->global_mem_cache_type = CL_NONE;
+  dev->local_mem_type = CL_GLOBAL;
+  dev->error_correction_support = CL_FALSE;
+  dev->host_unified_memory = CL_FALSE;
+
+  dev->available = CL_TRUE;
+  dev->compiler_available = CL_TRUE;
+  dev->spmd = CL_FALSE;
+  dev->workgroup_pass = CL_TRUE;
+  dev->execution_capabilities = CL_EXEC_KERNEL;
+  dev->queue_properties = CL_QUEUE_PROFILING_ENABLE;
+  dev->vendor = "TTA-Based Co-design Environment";
+  dev->profile = "EMBEDDED_PROFILE";
+  dev->extensions = TCE_DEVICE_EXTENSIONS;
+
+  dev->has_64bit_long = 1;
+  dev->autolocals_to_args = 1;
+
+  dev->global_as_id = TTA_ASID_GLOBAL;
+  dev->local_as_id = TTA_ASID_LOCAL;
+  dev->constant_as_id = TTA_ASID_CONSTANT;
+
+  SETUP_DEVICE_CL_VERSION (TCE_DEVICE_CL_VERSION_MAJOR,
+                           TCE_DEVICE_CL_VERSION_MINOR);
+
+  dev->parent_device = NULL;
+  // ttasim does not support partitioning
+  dev->max_sub_devices = 1;
+  dev->num_partition_properties = 1;
+  dev->partition_properties = (cl_device_partition_property *)calloc (
+      dev->num_partition_properties, sizeof (cl_device_partition_property));
+  dev->num_partition_types = 0;
+  dev->partition_type = NULL;
+
+  new TTASimDevice(dev, parameters);
   return CL_SUCCESS;
 }
 
