@@ -26,6 +26,7 @@
 #define POCL_CL_H
 
 #include "config.h"
+
 #include <assert.h>
 #include <stdio.h>
 #ifndef _MSC_VER
@@ -109,12 +110,17 @@ typedef pthread_mutex_t pocl_lock_t;
     }                                                                         \
   while (0)
 
-#define POCL_RELEASE_OBJECT(__OBJ__, __NEW_REFCOUNT__)                  \
-  do {                                                                  \
-    POCL_LOCK_OBJ (__OBJ__);                                            \
-    __NEW_REFCOUNT__ = --(__OBJ__)->pocl_refcount;                      \
-    POCL_UNLOCK_OBJ (__OBJ__);                                          \
-  } while (0)
+#define POCL_RELEASE_OBJECT_UNLOCKED(__OBJ__, __NEW_REFCOUNT__)               \
+  __NEW_REFCOUNT__ = --(__OBJ__)->pocl_refcount
+
+#define POCL_RELEASE_OBJECT(__OBJ__, __NEW_REFCOUNT__)                        \
+  do                                                                          \
+    {                                                                         \
+      POCL_LOCK_OBJ (__OBJ__);                                                \
+      POCL_RELEASE_OBJECT_UNLOCKED (__OBJ__, __NEW_REFCOUNT__);               \
+      POCL_UNLOCK_OBJ (__OBJ__);                                              \
+    }                                                                         \
+  while (0)
 
 #define POCL_RETAIN_OBJECT_UNLOCKED(__OBJ__)    \
     ++((__OBJ__)->pocl_refcount);
@@ -263,11 +269,9 @@ typedef struct pocl_argument_info {
   unsigned type_size;
 } pocl_argument_info;
 
+
 struct pocl_device_ops {
   const char *device_name;
-  void *shared_data; /* data to be shared by a devices of same type */
-  void (*init_device_infos) (unsigned j, struct _cl_device_id*);
-  /* implementation */
 
   /* New driver api extension for out-of-order execution and
      asynchronous devices.
@@ -275,7 +279,8 @@ struct pocl_device_ops {
      See basic and pthread driver for reference. */
 
   /* submit gives the command for the device. The command may be left in the cq
-     or stored to the device driver owning the cq. */
+     or stored to the device driver owning the cq. submit is called
+     with node->event locked, and must return with it unlocked. */
   void (*submit) (_cl_command_node *node, cl_command_queue cq);
 
   /* join is called by clFinish and this function blocks until all the enqueued
@@ -284,11 +289,12 @@ struct pocl_device_ops {
 
   /* flush is called when clFlush is called. This function ensures that
      commands will be eventually executed. It is up to the device what happens
-     here, if anything. See basic and pthread for reference. */
+     here, if anything. See basic and pthread for reference.*/
   void (*flush) (cl_device_id device, cl_command_queue cq);
 
   /* notify is used to communicate to a device driver that an event, it has
-     been waiting, has been completed. */
+     been waiting, has been completed. Upon call, both events are locked, and
+     must be locked also on return.*/
   void (*notify) (cl_device_id device, cl_event event, cl_event finished);
 
   /* broadcast is(has to be) called by the device driver when a command is
@@ -296,17 +302,22 @@ struct pocl_device_ops {
      It is used to broadcast notifications to device drivers waiting
      this event to complete.
      There is a default implementation for this. Use it if there is no need
-     to do anything special here. */
+     to do anything special here.
+     The default implementation calls notify(event, target_event) for the
+     list of events waiting on 'event'. */
   void (*broadcast) (cl_event event);
 
-  /* wait_event is blocking the execution until the waited event is complete.*/
+  /* wait_event is blocking the execution until the waited event is complete.
+     Called (and must return) with unlocked event. */
   void (*wait_event) (cl_device_id device, cl_event event);
 
   /* update_event is an alternative way of handling event status changes if
-     something device spesific needs to be done when the status of the event
+     something device specific needs to be done when the status of the event
      changes.
-     this function is called POCL_UPDATE_EVENT_* macros if available.
-     may be NULL, no need to implement if not needed. */
+     This function is called via POCL_UPDATE_EVENT_* macros if available.
+     All POCL_UPDATE_EVENT_ (except COMPLETE) must be called with LOCKED event.
+     may be NULL, no need to implement if not needed.
+     Called (and must return) with locked event. */
   void (*update_event) (cl_device_id device, cl_event event, cl_int status);
 
   /* free_event_data may be called when event is freed. Event data may only be
@@ -316,9 +327,7 @@ struct pocl_device_ops {
 
   /* /New driver api extension */
 
-  cl_int (*uninit) (cl_device_id device);
-  cl_int (*reinit) (cl_device_id device);
-
+  /* Detects & returns the number of available devices the driver finds on the system. */
   unsigned int (*probe) (struct pocl_device_ops *ops);
   /* Device initialization. Parameters:
    *  j : progressive index for the devices of the same type
@@ -326,15 +335,50 @@ struct pocl_device_ops {
    *  parameters : optional environment with device-specific parameters
    */
   cl_int (*init) (unsigned j, cl_device_id device, const char *parameters);
+  /* uninitializes the driver for a particular device. May free hardware resources. */
+  cl_int (*uninit) (cl_device_id device);
+  /* reinitializes the driver for a particular device. Called after uninit;
+   * the first initialization is done by 'init'. May be NULL */
+  cl_int (*reinit) (cl_device_id device);
+
+  /* if the driver needs to use hardware resources for command queues, use this */
   cl_int (*init_queue) (cl_command_queue queue);
   void (*free_queue) (cl_command_queue queue);
+
+  /* allocate a buffer in device memory */
   cl_int (*alloc_mem_obj) (cl_device_id device, cl_mem mem_obj, void* host_ptr);
   void *(*create_sub_buffer) (void *data, void* buffer, size_t origin, size_t size);
+  /* free a device buffer */
   void (*free) (cl_device_id device, cl_mem mem_obj);
-  void (*free_ptr) (cl_device_id device, void* mem_ptr);
-  void (*read) (void *data, void *host_ptr, const void *device_ptr,
-                size_t offset, size_t cb);
-  void (*read_rect) (void *data, void *host_ptr, void *device_ptr,
+
+  /* clEnqueSVMfree - free a SVM memory pointer. May be NULL if device doesn't
+   * support SVM. */
+  void (*svm_free) (cl_device_id dev, void *svm_ptr);
+  void *(*svm_alloc) (cl_device_id dev, cl_svm_mem_flags flags, size_t size);
+  void (*svm_map) (cl_device_id dev, void *svm_ptr);
+  void (*svm_unmap) (cl_device_id dev, void *svm_ptr);
+  /* we can use restrict here, because Spec says overlapping copy should return
+   * with CL_MEM_COPY_OVERLAP error. */
+  void (*svm_copy) (cl_device_id dev, void *__restrict__ dst,
+                    const void *__restrict__ src, size_t size);
+  void (*svm_fill) (cl_device_id dev, void *__restrict__ svm_ptr, size_t size,
+                    void *__restrict__ pattern, size_t pattern_size);
+
+  /* the following callbacks only deal with buffers (and IMAGE1D_BUFFER which
+   * is backed by a buffer), not images.  */
+
+  /* clEnqReadBuffer */
+  void (*read) (void *data,
+                void *__restrict__  dst_host_ptr,
+                pocl_mem_identifier * src_mem_id,
+                cl_mem src_buf,
+                size_t offset,
+                size_t size);
+  /* clEnqReadBufferRect */
+  void (*read_rect) (void *data,
+                     void *__restrict__ dst_host_ptr,
+                     pocl_mem_identifier * src_mem_id,
+                     cl_mem src_buf,
                      const size_t *buffer_origin,
                      const size_t *host_origin, 
                      const size_t *region,
@@ -342,9 +386,18 @@ struct pocl_device_ops {
                      size_t buffer_slice_pitch,
                      size_t host_row_pitch,
                      size_t host_slice_pitch);
-  void (*write) (void *data, const void *host_ptr, void *device_ptr, 
-                 size_t offset, size_t cb);
-  void (*write_rect) (void *data, const void *host_ptr, void *device_ptr,
+  /* clEnqWriteBuffer */
+  void (*write) (void *data,
+                 const void *__restrict__  src_host_ptr,
+                 pocl_mem_identifier * dst_mem_id,
+                 cl_mem dst_buf,
+                 size_t offset,
+                 size_t size);
+  /* clEnqWriteBufferRect */
+  void (*write_rect) (void *data,
+                      const void *__restrict__ src_host_ptr,
+                      pocl_mem_identifier * dst_mem_id,
+                      cl_mem dst_buf,
                       const size_t *buffer_origin,
                       const size_t *host_origin, 
                       const size_t *region,
@@ -352,57 +405,166 @@ struct pocl_device_ops {
                       size_t buffer_slice_pitch,
                       size_t host_row_pitch,
                       size_t host_slice_pitch);
-  void (*copy) (void *data, const void *src_ptr, size_t src_offset, 
-                void *__restrict__ dst_ptr, size_t dst_offset, size_t cb);
-  void (*copy_rect) (void *data, const void *src_ptr, void *dst_ptr,
+  /* clEnqCopyBuffer */
+  void (*copy) (void *data,
+                pocl_mem_identifier * dst_mem_id,
+                cl_mem dst_buf,
+                pocl_mem_identifier * src_mem_id,
+                cl_mem src_buf,
+                size_t dst_offset,
+                size_t src_offset,
+                size_t size);
+  /* clEnqCopyBufferRect */
+  void (*copy_rect) (void *data,
+                     pocl_mem_identifier * dst_mem_id,
+                     cl_mem dst_buf,
+                     pocl_mem_identifier * src_mem_id,
+                     cl_mem src_buf,
+                     const size_t *dst_origin,
                      const size_t *src_origin,
-                     const size_t *dst_origin, 
                      const size_t *region,
-                     size_t src_row_pitch,
-                     size_t src_slice_pitch,
                      size_t dst_row_pitch,
-                     size_t dst_slice_pitch);
+                     size_t dst_slice_pitch,
+                     size_t src_row_pitch,
+                     size_t src_slice_pitch);
 
-  void (*fill_rect) (void *data,
-                   void *__restrict__ const device_ptr,
-                   const size_t *__restrict__ const buffer_origin,
-                   const size_t *__restrict__ const region,
-                   size_t const buffer_row_pitch,
-                   size_t const buffer_slice_pitch,
-                   void *fill_pixel,
-                   size_t pixel_size);
-
-  void (*memfill) (void *ptr,
+  /* clEnqFillBuffer */
+  void (*memfill) (void *data,
+                   pocl_mem_identifier * dst_mem_id,
+                   cl_mem dst_buf,
                    size_t size,
                    size_t offset,
-                   const void* pattern,
+                   const void *__restrict__  pattern,
                    size_t pattern_size);
 
-  /* Maps 'size' bytes of device global memory at buf_ptr + offset to 
+  /* Maps 'size' bytes of device global memory at  + offset to
      host-accessible memory. This might or might not involve copying 
      the block from the device. */
-  void* (*map_mem) (void *data, void *buf_ptr, size_t offset, size_t size, void *host_ptr);
-  void* (*unmap_mem) (void *data, void *host_ptr, void *device_start_ptr, size_t offset, size_t size);
+  cl_int (*map_mem) (void *data,
+                     pocl_mem_identifier * src_mem_id,
+                     cl_mem src_buf,
+                     mem_mapping_t *map);
+  cl_int (*unmap_mem) (void *data,
+                       pocl_mem_identifier * dst_mem_id,
+                       cl_mem dst_buf,
+                       mem_mapping_t *map);
 
+  /* compile the fully linked LLVM IR to target-specific binaries. */
   void (*compile_kernel) (_cl_command_node* cmd, cl_kernel kernel, cl_device_id device);
+  /* clEnqueueNDRangeKernel */
   void (*run) (void *data, _cl_command_node* cmd);
+  /* for clEnqueueNativeKernel. may be NULL */
   void (*run_native) (void *data, _cl_command_node* cmd);
 
-  cl_ulong (*get_timer_value) (void *data); /* The current device timer value in nanoseconds. */
+  /* The current device timer value in nanoseconds. */
+  cl_ulong (*get_timer_value) (void *data);
 
   /* Perform initialization steps and can return additional
      build options that are required for the device. The caller
-     owns the returned string. */
+     owns the returned string. may be NULL */
   char* (*init_build) (void *data);
 
+  /* may be NULL */
   void (*init_target_machine) (void *data, void *target_machine);
 
+  /* returns a hash string that should identify the device. This string
+   * is used when writing/loading pocl binaries to decide compatibility. */
   char* (*build_hash) (cl_device_id device);
 
-    /* return supported image formats */
+  /* the following callbacks deal with images ONLY, with the exception of
+   * IMAGE1D_BUFFER type (which is implemented as a buffer).
+   * If the device does not support images, all of these may be NULL. */
+
+  /* return supported image formats */
   cl_int (*get_supported_image_formats) (cl_mem_flags flags,
                                          const cl_image_format **image_formats,
                                          cl_uint *num_image_formats);
+
+  /* returns a device specific pointer which may reference
+   * a hardware resource. May be NULL */
+  void* (*create_image) (void *data,
+                         const cl_image_format * image_format,
+                         const cl_image_desc *   image_desc,
+                         cl_mem image,
+                         cl_int *err);
+  /* free the device-specific pointer (image_data) from create_image() */
+  cl_int (*free_image) (void *data,
+                        cl_mem image,
+                        void *image_data);
+
+  /* creates a device-specific hardware resource for sampler. May be NULL */
+  void* (*create_sampler) (void *data,
+                           cl_sampler samp,
+                           cl_int *err);
+  cl_int (*free_sampler) (void *data,
+                          cl_sampler samp,
+                          void *sampler_data);
+
+  /* copies image to image, on the same device (or same global memory). */
+  cl_int (*copy_image_rect) (void *data,
+                             cl_mem src_image,
+                             cl_mem dst_image,
+                             pocl_mem_identifier *src_mem_id,
+                             pocl_mem_identifier *dst_mem_id,
+                             const size_t *src_origin,
+                             const size_t *dst_origin,
+                             const size_t *region);
+
+  /* copies a region from host OR device buffer to device image.
+   * clEnqueueCopyImageToBuffer: src_mem_id = buffer,
+   *     src_host_ptr = NULL, src_row_pitch = src_slice_pitch = 0
+   * clEnqueueWriteImage: src_mem_id = NULL,
+   *     src_host_ptr = host pointer, src_offset = 0
+   */
+  cl_int (*write_image_rect ) (void *data,
+                               cl_mem dst_image,
+                               pocl_mem_identifier *dst_mem_id,
+                               const void *__restrict__ src_host_ptr,
+                               pocl_mem_identifier *src_mem_id,
+                               const size_t *origin,
+                               const size_t *region,
+                               size_t src_row_pitch,
+                               size_t src_slice_pitch,
+                               size_t src_offset);
+
+  /* copies a region from device image to host or device buffer
+   * clEnqueueCopyBufferToImage: dst_mem_id = buffer,
+   *     dst_host_ptr = NULL, dst_row_pitch = dst_slice_pitch = 0
+   * clEnqueueReadImage: dst_mem_id = NULL,
+   *     dst_host_ptr = host pointer, dst_offset = 0
+   */
+  cl_int (*read_image_rect) (void *data,
+                             cl_mem src_image,
+                             pocl_mem_identifier *src_mem_id,
+                             void *__restrict__ dst_host_ptr,
+                             pocl_mem_identifier *dst_mem_id,
+                             const size_t *origin,
+                             const size_t *region,
+                             size_t dst_row_pitch,
+                             size_t dst_slice_pitch,
+                             size_t dst_offset);
+
+  /* maps the entire image from device to host */
+  cl_int (*map_image) (void *data,
+                       pocl_mem_identifier *mem_id,
+                       cl_mem src_image,
+                       mem_mapping_t *map);
+
+  /* unmaps the entire image from host to device */
+  cl_int (*unmap_image) (void *data,
+                         pocl_mem_identifier *mem_id,
+                         cl_mem dst_image,
+                         mem_mapping_t *map);
+
+  /* fill image with pattern */
+  cl_int (*fill_image)(void *data,
+                       cl_mem image,
+                       pocl_mem_identifier *mem_id,
+                       const size_t *origin,
+                       const size_t *region,
+                       const void *__restrict__ fill_pixel,
+                       size_t pixel_size);
+
 };
 
 typedef struct pocl_global_mem_t {
@@ -424,6 +586,12 @@ struct _cl_device_id {
   unsigned core_start;
   unsigned core_count;
   cl_uint max_work_item_dimensions;
+  /* when enabled, Workgroup LLVM pass will replace all printf() calls
+   * with calls to __pocl_printf and recursively change functions to
+   * add printf buffer arguments from pocl_context.
+   * Currently the pthread/basic devices require this; other devices
+   * implement printf their own way. */
+  int device_side_printf;
   size_t max_work_item_sizes[3];
   size_t max_work_group_size;
   size_t preferred_wg_size_multiple;
@@ -504,10 +672,8 @@ struct _cl_device_id {
   const char *profile;
   const char *version;
   const char *extensions;
-  cl_uint cl_version_major;    // 2
-  cl_uint cl_version_minor;    // 0
   const char *cl_version_std;  // "CL2.0"
-  cl_uint cl_version_int;      // 200
+  cl_ulong cl_version_int;     // 200
 
   void *data;
   const char* llvm_target_triplet; /* the llvm target triplet to use */
@@ -583,7 +749,6 @@ struct _cl_context {
   /* The device that should allocate SVM (might be == host)
    * NULL if none of devices in the context is SVM capable */
   cl_device_id svm_allocdev;
-
 };
 
 typedef struct _pocl_data_sync_item pocl_data_sync_item;
@@ -612,14 +777,6 @@ struct _cl_command_queue {
 };
 
 
-/* memory identifier: id to point the global memory where memory resides 
-                      + pointer to actual data */
-typedef struct _pocl_mem_identifier {
-  int available; /* ... in this mem objs context */
-  int global_mem_id;
-  void* mem_ptr;
-} pocl_mem_identifier;
-
 typedef struct _cl_mem cl_mem_t;
 struct _cl_mem {
   POCL_ICD_OBJECT
@@ -647,8 +804,6 @@ struct _cl_mem {
   cl_device_id shared_mem_allocation_owner;
   /* device where this mem obj resides */
   volatile cl_device_id owning_device;
-  /* latest event assosiated with the buffer, set NULL when event completed */
-  volatile cl_event latest_event;
   /* A linked list of regions of the buffer mapped to the 
      host memory */
   mem_mapping_t *mappings;
@@ -818,6 +973,7 @@ struct _cl_sampler {
   cl_bool             normalized_coords;
   cl_addressing_mode  addressing_mode;
   cl_filter_mode      filter_mode;
+  void**              device_data;
 };
 
 #define POCL_UPDATE_EVENT_QUEUED(__event)                                     \
@@ -971,9 +1127,9 @@ struct _cl_sampler {
 #define min(a,b) (((a) < (b)) ? (a) : (b))
 #define max(a,b) (((a) > (b)) ? (a) : (b))
 
-#ifndef __APPLE__
-  #include <endian.h>
-#else
+#endif
+
+#ifdef __APPLE__
   #include <libkern/OSByteOrder.h>
   #define htole16(x) OSSwapHostToLittleInt16(x)
   #define le16toh(x) OSSwapLittleToHostInt16(x)
@@ -981,8 +1137,10 @@ struct _cl_sampler {
   #define le32toh(x) OSSwapLittleToHostInt32(x)
   #define htole64(x) OSSwapHostToLittleInt64(x)
   #define le64toh(x) OSSwapLittleToHostInt64(x)
-#endif
-
+#elif defined(__FreeBSD__)
+  #include <sys/endian.h>
+#else
+  #include <endian.h>
 #endif
 
 #endif /* POCL_CL_H */
