@@ -69,7 +69,6 @@ using namespace llvm;
 
 POP_COMPILER_DIAGS
 
-
 /* Global pocl device to be used by passes if needed */
 cl_device_id currentPoclDevice = NULL;
 
@@ -277,7 +276,7 @@ int pocl_llvm_build_program(cl_program program,
   if (device->has_64bit_long)
     ss << "-Dcl_khr_int64 ";
 
-  ss << "-x cl ";
+  ss << "-xcl ";
   // Remove the inline keywords to force the user functions
   // to be included in the program. Otherwise they will
   // be removed and not inlined due to -O0.
@@ -399,11 +398,14 @@ int pocl_llvm_build_program(cl_program program,
   la->MathErrno = false; // -fno-math-errno
   la->NoBuiltin = true;  // -fno-builtin
   la->AsmBlocks = true;  // -fasm (?)
+
   la->setStackProtector(LangOptions::StackProtectorMode::SSPOff);
-  la->PICLevel = 0;
+
 #ifdef LLVM_OLDER_THAN_3_9
-  la->PIELevel = 0;
+  la->PICLevel = PICLevel::Large;
+  la->PIELevel = PICLevel::Large;
 #else
+  la->PICLevel = PICLevel::BigPIC;
   la->PIE = 0;
 #endif
 
@@ -533,15 +535,11 @@ int pocl_llvm_build_program(cl_program program,
 
   ++numberOfIRs;
 
-  // a workaround for errors with PIC + variables in constant addrspace
-  // fails on the test_convert_type_X tests with this error:
-  //   relocation R_X86_64_PC32 against symbol `char_values' can not be used
-  //   when making a shared object; recompile with -fPIC
 #ifdef LLVM_OLDER_THAN_3_9
-  (*mod)->setPICLevel(PICLevel::Default);
+  (*mod)->setPICLevel(PICLevel::Large);
 #else
-  (*mod)->setPICLevel(PICLevel::NotPIC);
-  (*mod)->setPIELevel(PIELevel::Default);
+  (*mod)->setPICLevel(PICLevel::BigPIC);
+  (*mod)->setPIELevel(PIELevel::Large);
 #endif
 
 
@@ -601,32 +599,46 @@ int pocl_llvm_link_program(cl_program program, unsigned device_i,
   llvm::Module **modptr = (llvm::Module **)&program->llvm_irs[device_i];
   int error;
 
+  currentPoclDevice = device;
+  llvm::Module *libmodule = kernel_library(device);
+  assert(libmodule != NULL);
+
   PoclCompilerMutexGuard lockHolder(NULL);
   InitializeLLVM();
 
   if (spir) {
 #ifdef ENABLE_SPIR
+
+    POCL_RETURN_ERROR_ON ((device->endian_little == CL_FALSE),
+                         CL_LINK_PROGRAM_FAILURE,
+                         "SPIR is only supported on little-endian devices\n");
+
     linked_module = parseModuleIRMem((char *)program->binaries[device_i],
                                      program->binary_sizes[device_i]);
 
-    Triple triple(device->llvm_target_triplet);
+    const std::string &spir_triple = linked_module->getTargetTriple();
+    size_t spir_addrbits = Triple(spir_triple).isArch64Bit() ? 64 : 32;
+
+    if (device->address_bits != spir_addrbits) {
+        delete linked_module;
+        POCL_RETURN_ERROR_ON (1, CL_LINK_PROGRAM_FAILURE,
+                       "Device address bits != SPIR binary triple address "
+                       "bits, device: %s / module: %s\n",
+                       device->llvm_target_triplet, spir_triple.c_str());
+    }
 
     /* Note this is a hack to get SPIR working. We'll be linking the
      * host kernel library (plain LLVM IR) to the SPIR program.bc,
      * so LLVM complains about incompatible DataLayouts.
      */
-    if (triple.getArch() == Triple::x86 || triple.getArch() == Triple::x86_64) {
-      if (linked_module->getTargetTriple() == std::string(SPIR_TARGET_TRIPLE)) {
-        linked_module->setTargetTriple(triple.getTriple());
-        linked_module->setDataLayout(SPIR_TARGET_DATALAYOUT);
-      } else {
-        POCL_MSG_ERR("Invalid SPIR binary triple\n");
-        return CL_LINK_PROGRAM_FAILURE;
-      }
-    } else {
-      POCL_MSG_ERR("SPIR is currently only supported on x86(-64)\n");
-      return CL_LINK_PROGRAM_FAILURE;
-    }
+    linked_module->setTargetTriple(libmodule->getTargetTriple());
+    linked_module->setDataLayout(libmodule->getDataLayout());
+#ifdef LLVM_OLDER_THAN_3_9
+    linked_module->setPICLevel(PICLevel::Large);
+#else
+    linked_module->setPICLevel(PICLevel::BigPIC);
+    linked_module->setPIELevel(PIELevel::Large);
+#endif
 
 #else
     POCL_MSG_ERR("SPIR not supported\n");
@@ -654,8 +666,10 @@ int pocl_llvm_link_program(cl_program program, unsigned device_i,
 #ifdef LLVM_OLDER_THAN_3_8
     if (Linker::LinkModules(mod, llvm::CloneModule(p))) {
       delete mod;
-#else
+#elif LLVM_OLDER_THAN_7_0
     if (Linker::linkModules(*mod, llvm::CloneModule(p))) {
+#else
+    if (Linker::linkModules(*mod, llvm::CloneModule(*p))) {
 #endif
       std::string msg = getDiagString();
       appendToProgramBuildLog(program, device_i, msg);
@@ -681,9 +695,6 @@ int pocl_llvm_link_program(cl_program program, unsigned device_i,
 
   if (!create_library) {
     // linked all the programs together, now link in the kernel library
-    currentPoclDevice = device;
-    llvm::Module *libmodule = kernel_library(device);
-    assert(libmodule != NULL);
     std::string log("Error(s) while linking: \n");
     if (link(linked_module, libmodule, log)) {
       appendToProgramBuildLog(program, device_i, log);
@@ -786,7 +797,7 @@ kernel_library
   const char *subdir = "host";
   bool is_host = true;
 #ifdef TCE_AVAILABLE
-  if (triple.getArch() == Triple::tce) {
+  if (triple.getArch() == Triple::tce || triple.getArch() == Triple::tcele) {
     subdir = "tce";
     is_host = false;
   }
@@ -817,49 +828,29 @@ kernel_library
 #ifdef ENABLE_POCL_BUILDING
   if (pocl_get_bool_option("POCL_BUILDING", 0)) {
     kernellib = BUILDDIR;
-#else
-  if (0) {
-#endif
     kernellib += "/lib/kernel/";
     kernellib += subdir;
-    // TODO: get this from the TCE target triplet
-    kernellib += "/kernel-";
-    kernellib += device->llvm_target_triplet;
-    if (is_host) {
-      kernellib += '-';
-      kernellib_fallback = kernellib;
-      kernellib_fallback += OCL_KERNEL_TARGET_CPU;
-      kernellib_fallback += ".bc";
-#ifdef KERNELLIB_HOST_DISTRO_VARIANTS
-      if (triple.getArch() == Triple::x86_64 ||
-          triple.getArch() == Triple::x86)
-        kernellib += getX86KernelLibName();
-      else
+  } else // POCL_BUILDING == 0, use install dir
 #endif
-        kernellib += device->llvm_cpu;
-    }
-  } else { // POCL_BUILDING == 0, use install dir
-    kernellib = POCL_INSTALL_PRIVATE_DATADIR;
-    kernellib += "/kernel-";
-    kernellib += device->llvm_target_triplet;
-    if (is_host) {
-      kernellib += '-';
-      kernellib_fallback = kernellib;
-      kernellib_fallback += OCL_KERNEL_TARGET_CPU;
-      kernellib_fallback += ".bc";
+  kernellib = POCL_INSTALL_PRIVATE_DATADIR;
+  kernellib += "/kernel-";
+  kernellib += device->llvm_target_triplet;
+  if (is_host) {
+    kernellib += '-';
+    kernellib_fallback = kernellib;
+    kernellib_fallback += OCL_KERNEL_TARGET_CPU;
+    kernellib_fallback += ".bc";
 #ifdef KERNELLIB_HOST_DISTRO_VARIANTS
-      if (triple.getArch() == Triple::x86_64 ||
-          triple.getArch() == Triple::x86)
-        kernellib += getX86KernelLibName();
-      else
+    if (triple.getArch() == Triple::x86_64 ||
+        triple.getArch() == Triple::x86)
+      kernellib += getX86KernelLibName();
+    else
 #endif
-        kernellib += device->llvm_cpu;
-    }
+      kernellib += device->llvm_cpu;
   }
   kernellib += ".bc";
 
   llvm::Module *lib;
-  SMDiagnostic Err;
 
   if (pocl_exists(kernellib.c_str()))
     {
