@@ -315,16 +315,15 @@ static int pocl_get_kernel_arg_function_metadata(const char *kernel_name,
                                                  llvm::Module *input,
                                                  cl_kernel kernel) {
   llvm::Function *Kernel = NULL;
-  int bitcode_is_spir = input->getTargetTriple().find("spir") == 0;
 
   // SPIR still uses the "opencl.kernels" MD.
-  if (bitcode_is_spir) {
-    auto status =
-      pocl_get_kernel_arg_module_metadata(kernel_name, input, kernel);
-    if (status == 0)
-      return status;
-    // Else go on, because it might be SPIR encoded with modern LLVM
+  llvm::NamedMDNode *opencl_kernels = input->getNamedMetadata("opencl.kernels");
+  int bitcode_is_old_spir = (opencl_kernels != NULL);
+
+  if (bitcode_is_old_spir) {
+    return pocl_get_kernel_arg_module_metadata(kernel_name, input, kernel);
   }
+  // Else go on, because it might be SPIR encoded with modern LLVM
 
   for (llvm::Module::iterator i = input->begin(), e = input->end();
        i != e; ++i) {
@@ -509,6 +508,8 @@ static int pocl_get_kernel_arg_function_metadata(const char *kernel_name,
     }
   }
   else {
+    POCL_MSG_WARN("no name metadata for kernel args"
+                  "(this is normal when compiling from SPIR) \n");
     // With SPIR 2.0 generated by more modern (3.9) Clang there is no more
     // "kernel_arg_name" metadata, so retrieve the name in another way
     // \todo Implement walking on the arguments themselves to get the names
