@@ -64,9 +64,6 @@
 #define PassManager legacy::PassManager
 #endif
 
-
-
-
 using namespace llvm;
 
 #include <string>
@@ -80,7 +77,11 @@ llvm::Module *parseModuleIR(const char *path) {
 
 void writeModuleIR(const Module *mod, std::string &str) {
   llvm::raw_string_ostream sos(str);
+#ifdef LLVM_OLDER_THAN_7_0
   WriteBitcodeToFile(mod, sos);
+#else
+  WriteBitcodeToFile(*mod, sos);
+#endif
   sos.str(); // flush
 }
 
@@ -120,7 +121,9 @@ int getModuleTriple(const char *input_stream, size_t size,
   return 0;
 }
 
-char *get_cpu_name() {
+char *
+get_llvm_cpu_name ()
+{
 
   StringRef r = llvm::sys::getHostCPUName();
 
@@ -192,10 +195,19 @@ void cpu_setup_vector_widths(cl_device_id dev) {
       VECWIDTH(cl_long);
   dev->native_vector_width_float = dev->preferred_vector_width_float =
       VECWIDTH(float);
+#ifdef _CL_DISABLE_DOUBLE
+  dev->native_vector_width_double = dev->preferred_vector_width_double = 0;
+#else
   dev->native_vector_width_double = dev->preferred_vector_width_double =
       VECWIDTH(double);
+#endif
+
+#ifdef _CL_DISABLE_HALF
+  dev->native_vector_width_half = dev->preferred_vector_width_half = 0;
+#else
   dev->native_vector_width_half = dev->preferred_vector_width_half =
       VECWIDTH(cl_short);
+#endif
 }
 
 int pocl_llvm_remove_file_on_signal(const char *file) {
