@@ -63,8 +63,6 @@ typedef struct list_item
   struct list_item *next;
 } list_item;
 
-
-
 void
 pocl_restore_ftz (unsigned ftz)
 {
@@ -222,18 +220,36 @@ pocl_size_ceil2(size_t x) {
   return ++x;
 }
 
-#ifndef HAVE_ALIGNED_ALLOC
+static void*
+pocl_memalign_alloc(size_t align_width, size_t size)
+{
+  void *ptr;
+  int status;
+
+#ifdef POCL_ANDROID
+  ptr = memalign (align_width, size);
+  return ptr;
+#elif defined(HAVE_POSIX_MEMALIGN)
+  status = posix_memalign (&ptr, align_width, size);
+  return ((status == 0) ? ptr : NULL);
+#elif (defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 201112L))
+  return aligned_alloc (align_width, size);
+#else
+#error Cannot find aligned malloc
+#endif
+}
+
 void *
 pocl_aligned_malloc (size_t alignment, size_t size)
 {
-# ifdef HAVE_POSIX_MEMALIGN
+#ifdef HAVE_ALIGNED_ALLOC
 
   /* make sure that size is a multiple of alignment, as posix_memalign
    * does not perform this test, whereas aligned_alloc does */
   if ((size & (alignment - 1)) != 0)
     {
-      errno = EINVAL;
-      return NULL;
+      size = size | (alignment - 1);
+      size += 1;
     }
 
   /* posix_memalign requires alignment to be at least sizeof(void *) */
@@ -251,8 +267,14 @@ pocl_aligned_malloc (size_t alignment, size_t size)
 
   return result;
 
-# else
+#else
+#error Cannot find aligned malloc
+#endif
 
+#if 0
+  /* this code works in theory, but there many places in pocl
+   * where aligned memory is used in the same pointers
+   * as memory allocated by other means */
   /* allow zero-sized allocations, force alignment to 1 */
   if (!size)
     alignment = 1;
@@ -280,15 +302,19 @@ pocl_aligned_malloc (size_t alignment, size_t size)
 
 #endif
 }
-#endif
 
-#if !defined HAVE_ALIGNED_ALLOC && !defined HAVE_POSIX_MEMALIGN
+#if 0
 void
 pocl_aligned_free (void *ptr)
 {
+#ifdef HAVE_ALIGNED_ALLOC
+  POCL_MEM_FREE (ptr);
+#else
+#error Cannot find aligned malloc
   /* extract pointer from original allocation and free it */
   if (ptr)
     free(*(void **)((uintptr_t)ptr - sizeof(void *)));
+#endif
 }
 #endif
 
@@ -490,6 +516,7 @@ cl_int pocl_create_command (_cl_command_node **cmd,
   return CL_SUCCESS;
 }
 
+/* call with node->event UNLOCKED */
 void pocl_command_enqueue (cl_command_queue command_queue,
                           _cl_command_node *node)
 {
@@ -530,10 +557,10 @@ void pocl_command_enqueue (cl_command_queue command_queue,
   /* node->event is unlocked by device_ops->submit */
 
   if (pocl_is_option_set("POCL_IMPLICIT_FINISH"))
-    POclFinish (command_queue);
+    POname(clFinish) (command_queue);
 }
 
-
+/* call (and return) with node->event locked */
 void
 pocl_command_push (_cl_command_node *node, 
                    _cl_command_node *volatile *ready_list, 
@@ -560,6 +587,7 @@ pocl_command_push (_cl_command_node *node,
     }
 }
 
+/* call with both event->queue and event UNLOCKED */
 int pocl_update_command_queue (cl_event event)
 {
   int cq_ready;
@@ -618,26 +646,6 @@ pocl_cl_mem_inherit_flags (cl_mem mem, cl_mem from_buffer, cl_mem_flags flags)
                | (from_buffer->flags & CL_MEM_COPY_HOST_PTR);
 }
 
-cl_int pocl_update_mem_obj_sync (cl_command_queue cq, _cl_command_node *cmd, 
-                                 cl_mem mem, char operation)
-{
-  int i;
-  POCL_LOCK_OBJ (mem);
-  mem->owning_device = cmd->device;
-  mem->latest_event = cmd->event;
-  POCL_UNLOCK_OBJ (mem);
-
-  for (i = 0; i < cmd->event->num_buffers; ++i)
-    {
-      if (cmd->event->mem_objs[i] == NULL)
-        {
-          cmd->event->mem_objs[i] = mem;
-          break;
-        }
-    }
-  return CL_SUCCESS;
-}
-
 int pocl_buffer_boundcheck(cl_mem buffer, size_t offset, size_t size) {
   POCL_RETURN_ERROR_ON ((offset > buffer->size), CL_INVALID_VALUE,
                         "offset(%zu) > buffer->size(%zu)\n", offset,
