@@ -29,6 +29,14 @@
 #include <string.h>
 #include "pocl_cl.h"
 
+#if defined(HAVE_POSIX_MEMALIGN) || defined(POCL_ANDROID) \
+     || (defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 201112L))
+#define HAVE_ALIGNED_ALLOC
+#else
+#error aligned malloc unavailable
+#endif
+
+
 #ifdef __GNUC__
 #pragma GCC visibility push(hidden)
 #endif
@@ -71,13 +79,9 @@ size_t pocl_size_ceil2(size_t x);
  * pointer must be freed with a call to pocl_aligned_free. Alignment
  * must be a non-zero power of 2.
  */
-#if defined HAVE_POSIX_MEMALIGN
+
 void *pocl_aligned_malloc(size_t alignment, size_t size);
-# define pocl_aligned_free free
-#else
-void *pocl_aligned_malloc(size_t alignment, size_t size);
-void pocl_aligned_free(void* ptr);
-#endif
+#define pocl_aligned_free(x) POCL_MEM_FREE(x)
 
 /* locks / unlocks two events in order of their event-id.
  * This avoids any potential deadlocks of threads should
@@ -145,10 +149,6 @@ pocl_update_command_queue (cl_event event);
 void pocl_cl_mem_inherit_flags (cl_mem mem, cl_mem from_buffer,
                                 cl_mem_flags flags);
 
-cl_int 
-pocl_update_mem_obj_sync (cl_command_queue cq, _cl_command_node *cmd, 
-                          cl_mem mem, char operation);
-
 void pocl_setup_context(cl_context context);
 
 /* Helpers for dealing with devices / subdevices */
@@ -265,9 +265,17 @@ float half_to_float (uint16_t value);
     }                                                                   \
   while (0)
 
-#define HANDLE_IMAGE1D_BUFFER(mem)                                            \
+#define IMAGE1D_TO_BUFFER(mem)                                                \
   mem = ((mem->is_image && (mem->type == CL_MEM_OBJECT_IMAGE1D_BUFFER))       \
              ? mem->buffer                                                    \
              : mem);
 
+#define IS_IMAGE1D_BUFFER(mem)                                                \
+  (mem->is_image && (mem->type == CL_MEM_OBJECT_IMAGE1D_BUFFER))
+
+#define IMAGE1D_ORIG_REG_TO_BYTES(mem, o, r)                                  \
+  size_t px = (mem->image_elem_size * mem->image_channels);                   \
+  size_t i1d_origin[3] = { o[0] * px, o[1], o[2] };                           \
+  size_t i1d_region[3] = { r[0] * px, r[1], r[2] };
+
 #endif
