@@ -112,6 +112,7 @@ double Sleef_exp2_u10 (double);
 double Sleef_exp10_u10 (double);
 double Sleef_expm1_u10 (double);
 double Sleef_log10_u10 (double);
+double Sleef_log2_u10 (double);
 double Sleef_log1p_u10 (double);
 Sleef_double2 Sleef_sincospi_u05 (double);
 Sleef_double2 Sleef_sincospi_u35 (double);
@@ -120,7 +121,9 @@ double Sleef_cospi_u05 (double);
 double Sleef_ldexp (double, int);
 int Sleef_ilogb (double);
 double Sleef_fma (double, double, double);
+double Sleef_sqrt (double);
 double Sleef_sqrt_u05 (double);
+double Sleef_sqrt_u35 (double);
 double Sleef_hypot_u05 (double, double);
 double Sleef_hypot_u35 (double, double);
 double Sleef_fabs (double);
@@ -176,6 +179,7 @@ float Sleef_exp2f_u10 (float);
 float Sleef_exp10f_u10 (float);
 float Sleef_expm1f_u10 (float);
 float Sleef_log10f_u10 (float);
+float Sleef_log2f_u10 (float);
 float Sleef_log1pf_u10 (float);
 Sleef_float2 Sleef_sincospif_u05 (float);
 Sleef_float2 Sleef_sincospif_u35 (float);
@@ -184,6 +188,7 @@ float Sleef_cospif_u05 (float d);
 float Sleef_ldexpf (float, int);
 int Sleef_ilogbf (float);
 float Sleef_fmaf (float, float, float);
+float Sleef_sqrtf (float);
 float Sleef_sqrtf_u05 (float);
 float Sleef_sqrtf_u35 (float);
 float Sleef_hypotf_u05 (float, float);
@@ -263,6 +268,7 @@ __m512d Sleef_exp2d8_u10_intrin (__m512d);
 __m512d Sleef_exp10d8_u10_intrin (__m512d);
 __m512d Sleef_expm1d8_u10_intrin (__m512d);
 __m512d Sleef_log10d8_u10_intrin (__m512d);
+__m512d Sleef_log2d8_u10_intrin (__m512d);
 __m512d Sleef_log1pd8_u10_intrin (__m512d);
 Sleef___m512d_2 Sleef_sincospid8_u05_intrin (__m512d);
 Sleef___m512d_2 Sleef_sincospid8_u35_intrin (__m512d);
@@ -271,6 +277,7 @@ __m512d Sleef_cospid8_u05_intrin (__m512d);
 __m512d Sleef_ldexpd8_intrin (__m512d, __m256i);
 __m256i Sleef_ilogbd8_intrin (__m512d);
 __m512d Sleef_fmad8_intrin (__m512d, __m512d, __m512d);
+__m512d Sleef_sqrtd8_intrin (__m512d);
 __m512d Sleef_sqrtd8_u05_intrin (__m512d);
 __m512d Sleef_sqrtd8_u35_intrin (__m512d);
 __m512d Sleef_hypotd8_u05_intrin (__m512d, __m512d);
@@ -337,6 +344,7 @@ __m512 Sleef_exp2f16_u10_intrin (__m512);
 __m512 Sleef_exp10f16_u10_intrin (__m512);
 __m512 Sleef_expm1f16_u10_intrin (__m512);
 __m512 Sleef_log10f16_u10_intrin (__m512);
+__m512 Sleef_log2f16_u10_intrin (__m512);
 __m512 Sleef_log1pf16_u10_intrin (__m512);
 Sleef___m512_2 Sleef_sincospif16_u05_intrin (__m512);
 Sleef___m512_2 Sleef_sincospif16_u35_intrin (__m512);
@@ -345,6 +353,7 @@ __m512 Sleef_cospif16_u05_intrin (__m512);
 __m512 Sleef_ldexpf16_intrin (__m512, __m512i);
 __m512i Sleef_ilogbf16_intrin (__m512);
 __m512 Sleef_fmaf16_intrin (__m512, __m512, __m512);
+__m512 Sleef_sqrtf16_intrin (__m512);
 __m512 Sleef_sqrtf16_u05_intrin (__m512);
 __m512 Sleef_sqrtf16_u35_intrin (__m512);
 __m512 Sleef_hypotf16_u05_intrin (__m512, __m512);
@@ -437,6 +446,7 @@ __m256d Sleef_exp2d4_u10_intrin (__m256d);
 __m256d Sleef_exp10d4_u10_intrin (__m256d);
 __m256d Sleef_expm1d4_u10_intrin (__m256d);
 __m256d Sleef_log10d4_u10_intrin (__m256d);
+__m256d Sleef_log2d4_u10_intrin (__m256d);
 __m256d Sleef_log1pd4_u10_intrin (__m256d);
 Sleef___m256d_2 Sleef_sincospid4_u05_intrin (__m256d);
 Sleef___m256d_2 Sleef_sincospid4_u35_intrin (__m256d);
@@ -445,6 +455,7 @@ __m256d Sleef_cospid4_u05_intrin (__m256d);
 __m256d Sleef_ldexpd4_intrin (__m256d, __m128i);
 __m128i Sleef_ilogbd4_intrin (__m256d);
 __m256d Sleef_fmad4_intrin (__m256d, __m256d, __m256d);
+__m256d Sleef_sqrtd4_intrin (__m256d);
 __m256d Sleef_sqrtd4_u05_intrin (__m256d);
 __m256d Sleef_sqrtd4_u35_intrin (__m256d);
 __m256d Sleef_hypotd4_u05_intrin (__m256d, __m256d);
@@ -511,6 +522,7 @@ __m256 Sleef_exp2f8_u10_intrin (__m256);
 __m256 Sleef_exp10f8_u10_intrin (__m256);
 __m256 Sleef_expm1f8_u10_intrin (__m256);
 __m256 Sleef_log10f8_u10_intrin (__m256);
+__m256 Sleef_log2f8_u10_intrin (__m256);
 __m256 Sleef_log1pf8_u10_intrin (__m256);
 Sleef___m256_2 Sleef_sincospif8_u05_intrin (__m256);
 Sleef___m256_2 Sleef_sincospif8_u35_intrin (__m256);
@@ -519,6 +531,7 @@ __m256 Sleef_cospif8_u05_intrin (__m256);
 __m256 Sleef_ldexpf8_intrin (__m256, __m256i);
 __m256i Sleef_ilogbf8_intrin (__m256);
 __m256 Sleef_fmaf8_intrin (__m256, __m256, __m256);
+__m256 Sleef_sqrtf8_intrin (__m256);
 __m256 Sleef_sqrtf8_u05_intrin (__m256);
 __m256 Sleef_sqrtf8_u35_intrin (__m256);
 __m256 Sleef_hypotf8_u05_intrin (__m256, __m256);
@@ -600,6 +613,7 @@ __m128d Sleef_exp2d2_u10_intrin (__m128d);
 __m128d Sleef_exp10d2_u10_intrin (__m128d);
 __m128d Sleef_expm1d2_u10_intrin (__m128d);
 __m128d Sleef_log10d2_u10_intrin (__m128d);
+__m128d Sleef_log2d2_u10_intrin (__m128d);
 __m128d Sleef_log1pd2_u10_intrin (__m128d);
 Sleef___m128d_2 Sleef_sincospid2_u05_intrin (__m128d);
 Sleef___m128d_2 Sleef_sincospid2_u35_intrin (__m128d);
@@ -608,6 +622,7 @@ __m128d Sleef_cospid2_u05_intrin (__m128d);
 __m128d Sleef_ldexpd2_intrin (__m128d, __m128i);
 __m128i Sleef_ilogbd2_intrin (__m128d);
 __m128d Sleef_fmad2_intrin (__m128d, __m128d, __m128d);
+__m128d Sleef_sqrtd2_intrin (__m128d);
 __m128d Sleef_sqrtd2_u05_intrin (__m128d);
 __m128d Sleef_sqrtd2_u35_intrin (__m128d);
 __m128d Sleef_hypotd2_u05_intrin (__m128d, __m128d);
@@ -674,6 +689,7 @@ __m128 Sleef_exp2f4_u10_intrin (__m128);
 __m128 Sleef_exp10f4_u10_intrin (__m128);
 __m128 Sleef_expm1f4_u10_intrin (__m128);
 __m128 Sleef_log10f4_u10_intrin (__m128);
+__m128 Sleef_log2f4_u10_intrin (__m128);
 __m128 Sleef_log1pf4_u10_intrin (__m128);
 Sleef___m128_2 Sleef_sincospif4_u05_intrin (__m128);
 Sleef___m128_2 Sleef_sincospif4_u35_intrin (__m128);
@@ -682,6 +698,7 @@ __m128 Sleef_cospif4_u05_intrin (__m128);
 __m128 Sleef_ldexpf4_intrin (__m128, __m128i);
 __m128i Sleef_ilogbf4_intrin (__m128);
 __m128 Sleef_fmaf4_intrin (__m128, __m128, __m128);
+__m128 Sleef_sqrtf4_intrin (__m128);
 __m128 Sleef_sqrtf4_u05_intrin (__m128);
 __m128 Sleef_sqrtf4_u35_intrin (__m128);
 __m128 Sleef_hypotf4_u05_intrin (__m128, __m128);
@@ -762,6 +779,7 @@ float32x4_t Sleef_exp2f4_u10_intrin (float32x4_t);
 float32x4_t Sleef_exp10f4_u10_intrin (float32x4_t);
 float32x4_t Sleef_expm1f4_u10_intrin (float32x4_t);
 float32x4_t Sleef_log10f4_u10_intrin (float32x4_t);
+float32x4_t Sleef_log2f4_u10_intrin (float32x4_t);
 float32x4_t Sleef_log1pf4_u10_intrin (float32x4_t);
 Sleef_float32x4_t_2 Sleef_sincospif4_u05_intrin (float32x4_t);
 Sleef_float32x4_t_2 Sleef_sincospif4_u35_intrin (float32x4_t);
@@ -771,6 +789,7 @@ float32x4_t Sleef_ldexpf4_intrin (float32x4_t, int32x4_t);
 int32x4_t Sleef_ilogbf4_intrin (float32x4_t);
 
 float32x4_t Sleef_fmaf4_intrin (float32x4_t, float32x4_t, float32x4_t);
+float32x4_t Sleef_sqrtf4_intrin (float32x4_t);
 float32x4_t Sleef_sqrtf4_u05_intrin (float32x4_t);
 float32x4_t Sleef_sqrtf4_u35_intrin (float32x4_t);
 float32x4_t Sleef_hypotf4_u05_intrin (float32x4_t, float32x4_t);
@@ -845,6 +864,7 @@ float64x2_t Sleef_exp2d2_u10_intrin (float64x2_t);
 float64x2_t Sleef_exp10d2_u10_intrin (float64x2_t);
 float64x2_t Sleef_expm1d2_u10_intrin (float64x2_t);
 float64x2_t Sleef_log10d2_u10_intrin (float64x2_t);
+float64x2_t Sleef_log2d2_u10_intrin (float64x2_t);
 float64x2_t Sleef_log1pd2_u10_intrin (float64x2_t);
 Sleef_float64x2_t_2 Sleef_sincospid2_u05_intrin (float64x2_t);
 Sleef_float64x2_t_2 Sleef_sincospid2_u35_intrin (float64x2_t);
@@ -854,6 +874,7 @@ float64x2_t Sleef_ldexpd2_intrin (float64x2_t, int32x2_t);
 int32x2_t Sleef_ilogbd2_intrin (float64x2_t);
 
 float64x2_t Sleef_fmad2_intrin (float64x2_t, float64x2_t, float64x2_t);
+float64x2_t Sleef_sqrtd2_intrin (float64x2_t);
 float64x2_t Sleef_sqrtd2_u05_intrin (float64x2_t);
 float64x2_t Sleef_sqrtd2_u35_intrin (float64x2_t);
 float64x2_t Sleef_hypotd2_u05_intrin (float64x2_t, float64x2_t);
