@@ -1,7 +1,7 @@
 // -*-C++-*-
 
-#ifndef VEC_AVX_FP16_16_H
-#define VEC_AVX_FP16_16_H
+#ifndef VEC_AVX_FP8_32_H
+#define VEC_AVX_FP8_32_H
 
 #include "floatprops.h"
 #include "mathfuncs.h"
@@ -14,13 +14,13 @@
 
 namespace vecmathlib {
 
-#define VECMATHLIB_HAVE_VEC_FP16_16
-template <> struct boolvec<fp16, 16>;
-template <> struct intvec<fp16, 16>;
-template <> struct realvec<fp16, 16>;
+#define VECMATHLIB_HAVE_VEC_FP8_32
+template <> struct boolvec<fp8, 32>;
+template <> struct intvec<fp8, 32>;
+template <> struct realvec<fp8, 32>;
 
-template <> struct boolvec<fp16, 16> : floatprops<fp16> {
-  static int const size = 16;
+template <> struct boolvec<fp8, 32> : floatprops<fp8> {
+  static int const size = 32;
   typedef bool scalar_t;
   typedef __m256i bvector_t;
   static int const alignment = sizeof(bvector_t);
@@ -56,15 +56,20 @@ public:
   // boolvec(boolvec const& x): v(x.v) {}
   // boolvec& operator=(boolvec const& x) { return v=x.v, *this; }
   boolvec(bvector_t x) : v(x) {}
-  boolvec(bool a) : v(_mm256_set1_epi16(from_bool(a))) {}
+  boolvec(bool a) : v(_mm256_set1_epi8(from_bool(a))) {}
   boolvec(bool const *as)
-      : v(_mm256_set_epi16(from_bool(as[15]), from_bool(as[14]),
-                           from_bool(as[13]), from_bool(as[12]),
-                           from_bool(as[11]), from_bool(as[10]),
-                           from_bool(as[9]), from_bool(as[8]), from_bool(as[7]),
-                           from_bool(as[6]), from_bool(as[5]), from_bool(as[4]),
-                           from_bool(as[3]), from_bool(as[2]), from_bool(as[1]),
-                           from_bool(as[0]))) {}
+      : v(_mm256_set_epi8(
+            from_bool(as[31]), from_bool(as[30]), from_bool(as[29]),
+            from_bool(as[28]), from_bool(as[27]), from_bool(as[26]),
+            from_bool(as[25]), from_bool(as[24]), from_bool(as[23]),
+            from_bool(as[22]), from_bool(as[21]), from_bool(as[20]),
+            from_bool(as[19]), from_bool(as[18]), from_bool(as[17]),
+            from_bool(as[16]), from_bool(as[15]), from_bool(as[14]),
+            from_bool(as[13]), from_bool(as[12]), from_bool(as[11]),
+            from_bool(as[10]), from_bool(as[9]), from_bool(as[8]),
+            from_bool(as[7]), from_bool(as[6]), from_bool(as[5]),
+            from_bool(as[4]), from_bool(as[3]), from_bool(as[2]),
+            from_bool(as[1]), from_bool(as[0]))) {}
 
   operator bvector_t() const { return v; }
   bool operator[](int n) const {
@@ -114,8 +119,8 @@ public:
   realvec_t ifthen(realvec_t x, realvec_t y) const; // defined after realvec
 };
 
-template <> struct intvec<fp16, 16> : floatprops<fp16> {
-  static int const size = 16;
+template <> struct intvec<fp8, 32> : floatprops<fp8> {
+  static int const size = 32;
   typedef int_t scalar_t;
   typedef __m256i ivector_t;
   static int const alignment = sizeof(ivector_t);
@@ -145,14 +150,17 @@ template <> struct intvec<fp16, 16> : floatprops<fp16> {
   // intvec(intvec const& x): v(x.v) {}
   // intvec& operator=(intvec const& x) { return v=x.v, *this; }
   intvec(ivector_t x) : v(x) {}
-  intvec(int_t a) : v(_mm256_set1_epi16(a)) {}
+  intvec(int_t a) : v(_mm256_set1_epi8(a)) {}
   intvec(int_t const *as)
-      : v(_mm256_set_epi16(as[15], as[14], as[13], as[12], as[11], as[10],
-                           as[9], as[8], as[7], as[6], as[5], as[4], as[3],
-                           as[2], as[1], as[0])) {}
+      : v(_mm256_set_epi8(as[31], as[30], as[29], as[28], as[27], as[26],
+                          as[25], as[24], as[23], as[22], as[21], as[20],
+                          as[19], as[18], as[17], as[16], as[15], as[14],
+                          as[13], as[12], as[11], as[10], as[9], as[8], as[7],
+                          as[6], as[5], as[4], as[3], as[2], as[1], as[0])) {}
   static intvec iota() {
-    return _mm256_set_epi16(15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
-                            0);
+    return _mm256_set_epi8(31, 30, 29, 28, 27, 26, 25, 24, 23, 22, 21, 20, 19,
+                           18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4,
+                           3, 2, 1, 0);
   }
 
   operator ivector_t() const { return v; }
@@ -184,27 +192,27 @@ template <> struct intvec<fp16, 16> : floatprops<fp16> {
 
   intvec operator+(intvec x) const {
 #ifdef __AVX2__
-    return _mm256_add_epi16(v, x.v);
+    return _mm256_add_epi8(v, x.v);
 #else
     __m128i vlo = _mm256_castsi256_si128(v);
     __m128i vhi = _mm256_extractf128_si256(v, 1);
     __m128i xvlo = _mm256_castsi256_si128(x.v);
     __m128i xvhi = _mm256_extractf128_si256(x.v, 1);
-    vlo = _mm_add_epi16(vlo, xvlo);
-    vhi = _mm_add_epi16(vhi, xvhi);
+    vlo = _mm_add_epi8(vlo, xvlo);
+    vhi = _mm_add_epi8(vhi, xvhi);
     return _mm256_insertf128_si256(_mm256_castsi128_si256(vlo), vhi, 1);
 #endif
   }
   intvec operator-(intvec x) const {
 #ifdef __AVX2__
-    return _mm256_sub_epi16(v, x.v);
+    return _mm256_sub_epi8(v, x.v);
 #else
     __m128i vlo = _mm256_castsi256_si128(v);
     __m128i vhi = _mm256_extractf128_si256(v, 1);
     __m128i xvlo = _mm256_castsi256_si128(x.v);
     __m128i xvhi = _mm256_extractf128_si256(x.v, 1);
-    vlo = _mm_sub_epi16(vlo, xvlo);
-    vhi = _mm_sub_epi16(vhi, xvhi);
+    vlo = _mm_sub_epi8(vlo, xvlo);
+    vhi = _mm_sub_epi8(vhi, xvhi);
     return _mm256_insertf128_si256(_mm256_castsi128_si256(vlo), vhi, 1);
 #endif
   }
@@ -245,34 +253,59 @@ template <> struct intvec<fp16, 16> : floatprops<fp16> {
 
   intvec lsr(int_t n) const {
 #ifdef __AVX2__
-    return _mm256_srli_epi16(v, n);
+    uint_t masklo = U(0x00ffU) >> U(n);
+    uint_t maskhi = U(0xff00U);
+    intvec mask = masklo | maskhi;
+    return intvec(_mm256_srai_epi16(v, n)) & mask;
 #else
     __m128i vlo = _mm256_castsi256_si128(v);
     __m128i vhi = _mm256_extractf128_si256(v, 1);
-    vlo = _mm_srli_epi16(vlo, n);
-    vhi = _mm_srli_epi16(vhi, n);
+    uint_t masklo = U(0x00ffU) >> U(n);
+    uint_t maskhi = U(0xff00U);
+    __m128i mask = _mm_set1_epi16(masklo | maskhi);
+    vlo = _mm_and_si128(_mm_srli_epi16(vlo, n), mask);
+    vhi = _mm_and_si128(_mm_srli_epi16(vhi, n), mask);
     return _mm256_insertf128_si256(_mm256_castsi128_si256(vlo), vhi, 1);
 #endif
   }
   intvec operator>>(int_t n) const {
 #ifdef __AVX2__
-    return _mm256_srai_epi16(v, n);
+    // There is no _mm256_srai_epi8. To emulate it, add 0x80 before
+    // shifting, and subtract the shifted 0x80 after shifting
+    intvec_t offset = (intvec_t)1 << (bits - 1);
+    return (*this + offset).lsr(n) - offset.lsr(n);
 #else
     __m128i vlo = _mm256_castsi256_si128(v);
     __m128i vhi = _mm256_extractf128_si256(v, 1);
-    vlo = _mm_srai_epi16(vlo, n);
-    vhi = _mm_srai_epi16(vhi, n);
+    uint_t masklo = U(0x00ffU);
+    uint_t maskhi = U(0xff00U);
+    __m128i vlolo = _mm_and_si128(_mm_srai_epi16(_mm_slli_epi16(vlo, 8), n + 8),
+                                  _mm_set1_epi16(masklo));
+    __m128i vlohi =
+        _mm_and_si128(_mm_srai_epi16(vlo, n), _mm_set1_epi16(maskhi));
+    vlo = _mm_or_si128(vlolo, vlohi);
+    __m128i vhilo = _mm_and_si128(_mm_srai_epi16(_mm_slli_epi16(vhi, 8), n + 8),
+                                  _mm_set1_epi16(masklo));
+    __m128i vhihi =
+        _mm_and_si128(_mm_srai_epi16(vhi, n), _mm_set1_epi16(maskhi));
+    vhi = _mm_or_si128(vhilo, vhihi);
     return _mm256_insertf128_si256(_mm256_castsi128_si256(vlo), vhi, 1);
 #endif
   }
   intvec operator<<(int_t n) const {
 #ifdef __AVX2__
-    return _mm256_slli_epi16(v, n);
+    uint_t masklo = U(0x00ffU);
+    uint_t maskhi = U(0xff00U) << U(n);
+    intvec mask = masklo | maskhi;
+    return intvec(_mm256_slli_epi16(v, n)) & mask;
 #else
     __m128i vlo = _mm256_castsi256_si128(v);
     __m128i vhi = _mm256_extractf128_si256(v, 1);
-    vlo = _mm_slli_epi16(vlo, n);
-    vhi = _mm_slli_epi16(vhi, n);
+    uint_t masklo = U(0x00ffU);
+    uint_t maskhi = U(0xff00U) << U(n);
+    __m128i mask = _mm_set1_epi16(masklo | maskhi);
+    vlo = _mm_and_si128(_mm_slli_epi16(vlo, n), mask);
+    vhi = _mm_and_si128(_mm_slli_epi16(vhi, n), mask);
     return _mm256_insertf128_si256(_mm256_castsi128_si256(vlo), vhi, 1);
 #endif
   }
@@ -280,61 +313,32 @@ template <> struct intvec<fp16, 16> : floatprops<fp16> {
   intvec &operator<<=(int_t n) { return *this = *this << n; }
 
   intvec lsr(intvec n) const {
-#ifdef __AVX2__
-    // TODO: Use permute instead of shift/mask?
-    __m256i mlo = _mm256_set1_epi32(U(0x0000ffff));
-    __m256i vlo = _mm256_and_si256(mlo, v);
-    __m256i vhi = v;
-    __m256i clo = _mm256_and_si256(mlo, n);
-    __m256i chi = _mm256_and_si256(mlo, _mm256_srli_epi32(n, 16));
-    __m256i rlo = _mm256_srlv_epi32(vlo, clo);
-    __m256i rhi = _mm256_andnot_si256(mlo, _mm256_srlv_epi32(vhi, chi));
-    return _mm256_or_si256(rhi, rlo);
-#else
     intvec r;
     for (int i = 0; i < size; ++i) {
       r.set_elt(i, U((*this)[i]) >> U(n[i]));
     }
     return r;
-#endif
   }
   intvec operator>>(intvec n) const {
-#ifdef __AVX2__
-    intvec_t offset = (intvec_t)1 << (bits - 1);
-    return (*this + offset).lsr(n) - offset.lsr(n);
-#else
     intvec r;
     for (int i = 0; i < size; ++i) {
       r.set_elt(i, (*this)[i] >> n[i]);
     }
     return r;
-#endif
   }
   intvec operator<<(intvec n) const {
-#ifdef __AVX2__
-    // TODO: Use permute instead of shift/mask?
-    __m256i mlo = _mm256_set1_epi32(U(0x0000ffff));
-    __m256i vlo = v;
-    __m256i vhi = _mm256_andnot_si256(mlo, v);
-    __m256i clo = _mm256_and_si256(mlo, n);
-    __m256i chi = _mm256_and_si256(mlo, _mm256_srli_epi32(n, 16));
-    __m256i rlo = _mm256_and_si256(mlo, _mm256_sllv_epi32(vlo, clo));
-    __m256i rhi = _mm256_sllv_epi32(vhi, chi);
-    return _mm256_or_si256(rhi, rlo);
-#else
     intvec r;
     for (int i = 0; i < size; ++i) {
       r.set_elt(i, (*this)[i] << n[i]);
     }
     return r;
-#endif
   }
   intvec &operator>>=(intvec n) { return *this = *this >> n; }
   intvec &operator<<=(intvec n) { return *this = *this << n; }
 
   boolvec_t operator==(intvec const &x) const {
 #ifdef __AVX2__
-    return _mm256_cmpeq_epi16(v, x.v);
+    return _mm256_cmpeq_epi8(v, x.v);
 #else
     return !(*this != x);
 #endif
@@ -348,7 +352,7 @@ template <> struct intvec<fp16, 16> : floatprops<fp16> {
   }
   boolvec_t operator<(intvec const &x) const {
 #ifdef __AVX2__
-    return _mm256_cmpgt_epi16(x.v, v);
+    return _mm256_cmpgt_epi8(x.v, v);
 #else
     // TODO: First compare sign; then if equal, compare sign of difference
     // TODO: Also look for intrinsics
@@ -369,17 +373,17 @@ template <> struct intvec<fp16, 16> : floatprops<fp16> {
   intvec_t min(intvec_t x) const;
 };
 
-template <> struct realvec<fp16, 16> : floatprops<fp16> {
-  static int const size = 16;
+template <> struct realvec<fp8, 32> : floatprops<fp8> {
+  static int const size = 32;
   typedef real_t scalar_t;
   typedef __m256i vector_t;
   static int const alignment = sizeof(vector_t);
 
   static char const *name() {
 #ifdef __AVX2__
-    return "<AVX2:16*fp16>";
+    return "<AVX2:32*fp8>";
 #else
-    return "<AVX:16*fp16>";
+    return "<AVX:32*fp8>";
 #endif
   }
   void barrier() { __asm__("" : "+x"(v)); }
@@ -409,15 +413,20 @@ template <> struct realvec<fp16, 16> : floatprops<fp16> {
   // realvec(realvec const& x): v(x.v) {}
   // realvec& operator=(realvec const& x) { return v=x.v, *this; }
   realvec(vector_t x) : v(x) {}
-  realvec(real_t a) : v(_mm256_set1_epi16(FP::as_int(a))) {}
+  realvec(real_t a) : v(_mm256_set1_epi8(FP::as_int(a))) {}
   realvec(real_t const *as)
-      : v(_mm256_set_epi16(
-            FP::as_int(as[15]), FP::as_int(as[14]), FP::as_int(as[13]),
-            FP::as_int(as[12]), FP::as_int(as[11]), FP::as_int(as[10]),
-            FP::as_int(as[9]), FP::as_int(as[8]), FP::as_int(as[7]),
-            FP::as_int(as[6]), FP::as_int(as[5]), FP::as_int(as[4]),
-            FP::as_int(as[3]), FP::as_int(as[2]), FP::as_int(as[1]),
-            FP::as_int(as[0]))) {}
+      : v(_mm256_set_epi8(
+            FP::as_int(as[31]), FP::as_int(as[30]), FP::as_int(as[29]),
+            FP::as_int(as[28]), FP::as_int(as[27]), FP::as_int(as[26]),
+            FP::as_int(as[25]), FP::as_int(as[24]), FP::as_int(as[23]),
+            FP::as_int(as[22]), FP::as_int(as[21]), FP::as_int(as[20]),
+            FP::as_int(as[19]), FP::as_int(as[18]), FP::as_int(as[17]),
+            FP::as_int(as[16]), FP::as_int(as[15]), FP::as_int(as[14]),
+            FP::as_int(as[13]), FP::as_int(as[12]), FP::as_int(as[11]),
+            FP::as_int(as[10]), FP::as_int(as[9]), FP::as_int(as[8]),
+            FP::as_int(as[7]), FP::as_int(as[6]), FP::as_int(as[5]),
+            FP::as_int(as[4]), FP::as_int(as[3]), FP::as_int(as[2]),
+            FP::as_int(as[1]), FP::as_int(as[0]))) {}
 
   operator vector_t() const { return v; }
   real_t operator[](int n) const {
@@ -545,51 +554,50 @@ template <> struct realvec<fp16, 16> : floatprops<fp16> {
 
 // boolvec definitions
 
-inline intvec<fp16, 16> boolvec<fp16, 16>::as_int() const { return v; }
+inline intvec<fp8, 32> boolvec<fp8, 32>::as_int() const { return v; }
 
-inline intvec<fp16, 16> boolvec<fp16, 16>::convert_int() const {
+inline intvec<fp8, 32> boolvec<fp8, 32>::convert_int() const {
   return lsr(as_int(), bits - 1);
 }
 
-inline boolvec<fp16, 16> boolvec<fp16, 16>::ifthen(boolvec_t x,
-                                                   boolvec_t y) const {
+inline boolvec<fp8, 32> boolvec<fp8, 32>::ifthen(boolvec_t x,
+                                                 boolvec_t y) const {
   return ifthen(x.as_int(), y.as_int()).as_bool();
 }
 
-inline intvec<fp16, 16> boolvec<fp16, 16>::ifthen(intvec_t x,
-                                                  intvec_t y) const {
+inline intvec<fp8, 32> boolvec<fp8, 32>::ifthen(intvec_t x, intvec_t y) const {
   return ((-convert_int() & x) | (~ - convert_int() & y));
 }
 
-inline realvec<fp16, 16> boolvec<fp16, 16>::ifthen(realvec_t x,
-                                                   realvec_t y) const {
+inline realvec<fp8, 32> boolvec<fp8, 32>::ifthen(realvec_t x,
+                                                 realvec_t y) const {
   return ifthen(x.as_int(), y.as_int()).as_float();
 }
 
 // intvec definitions
 
-inline intvec<fp16, 16> intvec<fp16, 16>::abs() const {
+inline intvec<fp8, 32> intvec<fp8, 32>::abs() const {
 #ifdef __AVX2__
-  return _mm256_abs_epi16(v);
+  return _mm256_abs_epi8(v);
 #else
   return MF::vml_abs(*this);
 #endif
 }
 
-inline realvec<fp16, 16> intvec<fp16, 16>::as_float() const { return v; }
+inline realvec<fp8, 32> intvec<fp8, 32>::as_float() const { return v; }
 
-inline realvec<fp16, 16> intvec<fp16, 16>::convert_float() const {
+inline realvec<fp8, 32> intvec<fp8, 32>::convert_float() const {
   __builtin_unreachable();
 }
 
-inline intvec<fp16, 16> intvec<fp16, 16>::max(intvec_t x) const {
+inline intvec<fp8, 32> intvec<fp8, 32>::max(intvec_t x) const {
   return MF::vml_max(*this, x);
 }
 
-inline intvec<fp16, 16> intvec<fp16, 16>::min(intvec_t x) const {
+inline intvec<fp8, 32> intvec<fp8, 32>::min(intvec_t x) const {
   return MF::vml_min(*this, x);
 }
 
 } // namespace vecmathlib
 
-#endif // #ifndef VEC_AVX_FP16_16_H
+#endif // #ifndef VEC_AVX_FP8_32_H
