@@ -93,7 +93,11 @@ BarrierTailReplication::runOnFunction(Function &F)
 
   bool changed = ProcessFunction(F);
 
+
+  // In LLVM 7+, it is replaced by some assert(verify()) in LLVM itself
+#ifdef LLVM_OLDER_THAN_7_0
   DT->verifyDomTree();
+#endif
 
   LI->verifyAnalysis();
   /* The created tails might contain PHI nodes with operands 
@@ -375,7 +379,9 @@ BarrierTailReplication::ReplicateBasicBlocks(BasicBlockVector &new_graph,
           
           // Get value for original incoming edge and add new predicate.
           Value *v = phi->getIncomingValueForBlock(b);
-          Value *new_v = reference_map[v];
+          Value *new_v = reference_map.find(v) == reference_map.end() ?
+            NULL : reference_map[v];
+
           if (new_v == NULL) {
             /* This case can happen at least when replicating a latch 
                block in a b-loop. The value produced might be from a common
@@ -413,8 +419,13 @@ BarrierTailReplication::UpdateReferences(const BasicBlockVector &graph,
     for (BasicBlock::iterator i2 = b->begin(), e2 = b->end();
          i2 != e2; ++i2) {
       Instruction *i = &*i2;
+#ifdef LLVM_OLDER_THAN_3_9
       RemapInstruction(i, reference_map,
                        RF_IgnoreMissingEntries | RF_NoModuleLevelChanges);
+#else
+      RemapInstruction(i, reference_map,
+                       RF_IgnoreMissingLocals | RF_NoModuleLevelChanges);
+#endif
     }
   }
 }
